# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>
Copyright (C) 2008 Rob McMullen <rob.mcmullen@gmail.com>
Copyright (C) 2008 Thomas Sonne Olesen <tpo@sonnet.dk>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import command, widgets, domain
from taskcoachlib.domain import note
from taskcoachlib.gui import uicommand, menu, dialog
from taskcoachlib.i18n import _
import base
import mixin
import inplace_editor
import wx


class BaseNoteViewer(mixin.AttachmentDropTargetMixin,  # pylint: disable=W0223
                     mixin.SearchableViewerMixin, 
                     mixin.SortableViewerForNotesMixin,
                     mixin.AttachmentColumnMixin, 
                     base.CategorizableViewerMixin,
                     base.WithAttachmentsViewerMixin,
                     base.SortableViewerWithColumns, base.TreeViewer):
    SorterClass = note.NoteSorter
    defaultTitle = _('Notes')
    defaultBitmap = 'note_icon'
    
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('settingsSection', 'noteviewer')
        self.notesToShow = kwargs.get('notesToShow', None)
        super(BaseNoteViewer, self).__init__(*args, **kwargs)
        for eventType in (note.Note.appearanceChangedEventType(), 
                          note.Note.subjectChangedEventType()):
            self.registerObserver(self.onAttributeChanged_Deprecated, 
                                  eventType)

    def domainObjectsToView(self):
        return self.taskFile.notes() if self.notesToShow is None else self.notesToShow

    def curselectionIsInstanceOf(self, class_):
        return class_ == note.Note

    def createWidget(self):
        imageList = self.createImageList()  # Has side-effects
        self._columns = self._createColumns()
        itemPopupMenu = menu.NotePopupMenu(self.parent, self.settings,
            self.taskFile.categories(), self)
        columnPopupMenu = menu.ColumnPopupMenu(self)
        self._popupMenus.extend([itemPopupMenu, columnPopupMenu])
        widget = widgets.TreeListCtrl(self, self.columns(), self.onSelect,
            uicommand.Edit(viewer=self),
            uicommand.NoteDragAndDrop(viewer=self, notes=self.presentation()),
            itemPopupMenu, columnPopupMenu,
            resizeableColumn=1 if self.hasOrderingColumn() else 0,
            validateDrag=self.validateDrag,
            **self.widgetCreationKeywordArguments())
        if self.hasOrderingColumn():
            widget.SetMainColumn(1)
        widget.AssignImageList(imageList)  # pylint: disable=E1101
        return widget
    
    def createFilter(self, notes):
        notes = super(BaseNoteViewer, self).createFilter(notes)
        return domain.base.DeletedFilter(notes)
    
    def createCreationToolBarUICommands(self):
        return (uicommand.NoteNew(notes=self.presentation(),
                                  settings=self.settings, viewer=self),
                uicommand.NewSubItem(viewer=self),) + \
            super(BaseNoteViewer, self).createCreationToolBarUICommands()
        
    def createColumnUICommands(self):
        return [\
            uicommand.ToggleAutoColumnResizing(viewer=self,
                                               settings=self.settings),
            None,
            uicommand.ViewColumn(menuText=_('&Manual ordering'),
                helpText=_('Show/hide the manual ordering column'),
                setting='ordering', viewer=self),
            uicommand.ViewColumn(menuText=_('&Description'),
                helpText=_('Show/hide description column'),
                setting='description', viewer=self),
            uicommand.ViewColumn(menuText=_('&Attachments'),
                helpText=_('Show/hide attachments column'),
                setting='attachments', viewer=self),
            uicommand.ViewColumn(menuText=_('&Categories'),
                helpText=_('Show/hide categories column'),
                setting='categories', viewer=self),
            uicommand.ViewColumn(menuText=_('&Creation date'),
                helpText=_('Show/hide creation date column'),
                setting='creationDateTime', viewer=self),
            uicommand.ViewColumn(menuText=_('&Modification date'),
                helpText=_('Show/hide last modification date column'),
                setting='modificationDateTime', viewer=self)]

    def _createColumns(self):
        orderingColumn = widgets.Column('ordering', u'',
            width=self.getColumnWidth('ordering'),
            resizeCallback=self.onResizeColumn,
            renderCallback=lambda note: '',
            imageIndicesCallback=self.orderingImageIndices,
            sortCallback=uicommand.ViewerSortByCommand(viewer=self,
                value='ordering', menuText=_('&Manual ordering'),
                helpText=_('Sort notes manually')))
        # XXXCHECK editCallback & co
        subjectColumn = widgets.Column('subject', _('Subject'), 
            width=self.getColumnWidth('subject'), 
            resizeCallback=self.onResizeColumn,
            renderCallback=lambda note: note.subject(),
            sortCallback=uicommand.ViewerSortByCommand(viewer=self, 
                value='subject', menuText=_('&Subject'), 
                helpText=_('Sort notes by subject')),
            imageIndicesCallback=self.subjectImageIndices,
            editCallback=self.onEditSubject,
            editControl=inplace_editor.SubjectCtrl)
        descriptionColumn = widgets.Column('description', _('Description'),
            note.Note.descriptionChangedEventType(),
            width=self.getColumnWidth('description'), 
            resizeCallback=self.onResizeColumn,
            renderCallback=lambda note: note.description(),
            sortCallback=uicommand.ViewerSortByCommand(viewer=self, 
                value='description', menuText=_('&Description'), 
                helpText=_('Sort notes by description')),
            editCallback=self.onEditDescription,
            editControl=inplace_editor.DescriptionCtrl)
        attachmentsColumn = widgets.Column('attachments', '', 
            note.Note.attachmentsChangedEventType(),  # pylint: disable=E1101
            width=self.getColumnWidth('attachments'),
            alignment=wx.LIST_FORMAT_LEFT,
            imageIndicesCallback=self.attachmentImageIndices,  # pylint: disable=E1101
            headerImageIndex=self.imageIndex['paperclip_icon'],
            renderCallback=lambda note: '')
        categoriesColumn = widgets.Column('categories', _('Categories'),
            note.Note.categoryAddedEventType(), 
            note.Note.categoryRemovedEventType(), 
            note.Note.categorySubjectChangedEventType(),
            note.Note.expansionChangedEventType(),
            width=self.getColumnWidth('categories'),
            resizeCallback=self.onResizeColumn,
            renderCallback=self.renderCategories,
            sortCallback=uicommand.ViewerSortByCommand(viewer=self, 
                value='categories', menuText=_('&Categories'), 
                helpText=_('Sort notes by categories')))
        creationDateTimeColumn = widgets.Column('creationDateTime', 
            _('Creation date'), width=self.getColumnWidth('creationDateTime'),
            resizeCallback=self.onResizeColumn,
            renderCallback=self.renderCreationDateTime,
            sortCallback=uicommand.ViewerSortByCommand(viewer=self,
                value='creationDateTime', menuText=_('&Creation date'),
                helpText=_('Sort notes by creation date')))
        modificationDateTimeColumn = widgets.Column('modificationDateTime', 
            _('Modification date'), 
            width=self.getColumnWidth('modificationDateTime'),
            resizeCallback=self.onResizeColumn,
            renderCallback=self.renderModificationDateTime,
            sortCallback=uicommand.ViewerSortByCommand(viewer=self,
                value='modificationDateTime', menuText=_('&Modification date'),
                helpText=_('Sort notes by last modification date')),
            *note.Note.modificationEventTypes())
        return [orderingColumn, subjectColumn, descriptionColumn, attachmentsColumn, 
                categoriesColumn, creationDateTimeColumn, 
                modificationDateTimeColumn]
                    
    def isShowingNotes(self):
        return True

    def statusMessages(self):
        status1 = _('Notes: %d selected, %d total') % \
            (len(self.curselection()), len(self.presentation()))
        status2 = _('Status: n/a')
        return status1, status2

    def newItemDialog(self, *args, **kwargs):
        kwargs['categories'] = self.taskFile.categories().filteredCategories()
        return super(BaseNoteViewer, self).newItemDialog(*args, **kwargs)
    
    def deleteItemCommand(self):
        return command.DeleteNoteCommand(self.presentation(), 
            self.curselection(), 
            shadow=self.settings.getboolean('feature', 'syncml'))
        
    def itemEditorClass(self):
        return dialog.editor.NoteEditor

    def newItemCommandClass(self):
        return command.NewNoteCommand
    
    def newSubItemCommandClass(self):
        return command.NewSubNoteCommand


class NoteViewer(mixin.FilterableViewerForCategorizablesMixin, BaseNoteViewer):  # pylint: disable=W0223
    pass



#!/usr/bin/env python
"""
Test the various hashing methods

This test runs with the gntp.config module so that we can
get away without having to hardcode our password in a test
script. Please fill out your ~/.gntp config before running
"""
import os
from gntp.test import GNTPTestCase
from gntp import UnsupportedError


class TestHash(GNTPTestCase):

	def _hash(self, hashName):
		self.growl.passwordHash = hashName
		return self._notify(description=hashName)

	def test_config(self):
		"""Test to see if our config file exists

		If our config file doesn't exist, then we have no
		password to test with, so our password hash is no good
		"""
		config = os.path.expanduser('~/.gntp')
		self.assertIsTrue(os.path.exists(config))

	def test_md5(self):
		self.assertIsTrue(self._hash('MD5'))

	def test_sha1(self):
		self.assertIsTrue(self._hash('SHA1'))

	def test_sha256(self):
		self.assertIsTrue(self._hash('SHA256'))

	def test_sha512(self):
		self.assertIsTrue(self._hash('SHA512'))

	def test_fake(self):
		'''Fake hash should not work'''
		self.assertRaises(UnsupportedError, self._hash, 'fake-hash')



# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from .. import sessiontempfile  # pylint: disable=F0401
from taskcoachlib import meta, patterns
from taskcoachlib.changes import ChangeMonitor
from taskcoachlib.domain import base, date, effort, task, category, categorizable, note, attachment
from taskcoachlib.i18n import translate
from taskcoachlib.syncml.config import SyncMLConfigNode, createDefaultSyncConfig
from taskcoachlib.thirdparty.deltaTime import nlTimeExpression
from taskcoachlib.thirdparty.guid import generate
import StringIO
import os
import re
import stat
import wx
import xml.etree.ElementTree as ET


def parseAndAdjustDateTime(string, *timeDefaults):
    dateTime = date.parseDateTime(string, *timeDefaults)
    if dateTime != date.DateTime() and dateTime is not None and \
        dateTime.time() == date.Time(23, 59, 0, 0):
        dateTime = date.DateTime(year=dateTime.year,
                                 month=dateTime.month,
                                 day=dateTime.day,
                                 hour=23, minute=59, second=59, microsecond=999999)
    return dateTime


class PIParser(ET.XMLTreeBuilder):
    """See http://effbot.org/zone/element-pi.htm"""
    def __init__(self):
        ET.XMLTreeBuilder.__init__(self)
        self._parser.ProcessingInstructionHandler = self.handle_pi
        self.tskversion = meta.data.tskversion

    def handle_pi(self, target, data):
        if target == 'taskcoach':
            match_object = re.search('tskversion="(\d+)"', data)
            self.tskversion = int(match_object.group(1))


class XMLReaderTooNewException(Exception):
    pass


class XMLReader(object):
    ''' Class for reading task files in the default XML task file format. '''
    defaultStartTime = (0, 0, 0, 0)
    defaultEndTime = (23, 59, 59, 999999)

    def __init__(self, fd):
        self.__fd = fd
        self.__default_font_size = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT).GetPointSize()
        self.__modification_datetimes = {}
        self.__prerequisites = {}
        self.__categorizables = {}

    def tskversion(self):
        ''' Return the version of the current task file. Note that this is not
            the version of the application. The task file has its own version
            numbering (a number that is increasing on every change). '''
        return self.__tskversion

    def read(self):
        ''' Read the task file and return the tasks, categories, notes, SyncML
            configuration and GUID. '''
        if self.__has_broken_lines():
            self.__fix_broken_lines()
        parser = PIParser()
        tree = ET.parse(self.__fd, parser)
        root = tree.getroot()
        self.__tskversion = parser.tskversion  # pylint: disable=W0201
        if self.__tskversion > meta.data.tskversion:
            # Version number of task file is too high
            raise XMLReaderTooNewException
        tasks = self.__parse_task_nodes(root)
        self.__resolve_prerequisites_and_dependencies(tasks)
        notes = self.__parse_note_nodes(root)
        if self.__tskversion <= 13:
            categories = self.__parse_category_nodes_from_task_nodes(root)
        else:
            categories = self.__parse_category_nodes(root)
        self.__resolve_categories(categories, tasks, notes)

        guid = self.__parse_guid_node(root.find('guid'))
        syncml_config = self.__parse_syncml_node(root, guid)
        
        for object, modification_datetime in self.__modification_datetimes.iteritems():
            object.setModificationDateTime(modification_datetime)

        changesName = self.__fd.name + '.delta'
        if os.path.exists(changesName):
            changes = ChangesXMLReader(file(self.__fd.name + '.delta', 'rU')).read()
        else:
            changes = dict()

        return tasks, categories, notes, syncml_config, changes, guid

    def __has_broken_lines(self):
        ''' tskversion 24 may contain newlines in element tags. '''
        has_broken_lines = '><spds><sources><TaskCoach-\n' in self.__fd.read()
        self.__fd.seek(0)
        return has_broken_lines
    
    def __fix_broken_lines(self):
        ''' Remove spurious newlines from element tags. '''
        self.__origFd = self.__fd  # pylint: disable=W0201
        self.__fd = StringIO.StringIO()
        self.__fd.name = self.__origFd.name
        lines = self.__origFd.readlines()
        for index in xrange(len(lines)):
            if lines[index].endswith('<TaskCoach-\n') or \
               lines[index].endswith('</TaskCoach-\n'):
                lines[index] = lines[index][:-1]  # Remove newline
                lines[index + 1] = lines[index + 1][:-1]  # Remove newline
        self.__fd.write(''.join(lines))
        self.__fd.seek(0)

    def __parse_task_nodes(self, node):
        ''' Recursively parse all tasks from the node and return a list of
            task instances. '''
        return [self._parse_task_node(child) for child in node.findall('task')]
                
    def __resolve_prerequisites_and_dependencies(self, tasks):
        ''' Replace all prerequisites with the actual task instances
            and set the dependencies. '''
        tasks_by_id = dict()
        
        def collect_ids(tasks):
            ''' Create a mapping from task ids to task instances. '''
            for each_task in tasks:
                tasks_by_id[each_task.id()] = each_task
                collect_ids(each_task.children())
        
        def resolve_ids(tasks):
            ''' Replace all prerequisites ids with actual task instances and 
                set the dependencies. '''
            for each_task in tasks:
                if each_task.isDeleted():
                    # Don't restore prerequisites and dependencies for deleted
                    # tasks
                    for deleted_task in [each_task] + \
                                        each_task.children(recursive=True):
                        deleted_task.setPrerequisites([])
                    continue
                prerequisites = set()
                for prerequisiteId in self.__prerequisites.get(each_task.id(), []):
                    try:
                        prerequisites.add(tasks_by_id[prerequisiteId])
                    except KeyError:
                        # Release 1.2.11 and older have a bug where tasks can
                        # have prerequisites listed that don't exist anymore
                        pass
                each_task.setPrerequisites(prerequisites)
                for prerequisite in prerequisites:
                    prerequisite.addDependencies([each_task])
                resolve_ids(each_task.children())
                
        collect_ids(tasks)
        resolve_ids(tasks)

    def __resolve_categories(self, categories, tasks, notes):
        def mapCategorizables(obj, resultMap, categoryMap):
            if isinstance(obj, categorizable.CategorizableCompositeObject):
                resultMap[obj.id()] = obj
            if isinstance(obj, category.Category):
                categoryMap[obj.id()] = obj
            if isinstance(obj, base.CompositeObject):
                for child in obj.children():
                    mapCategorizables(child, resultMap, categoryMap)
            if isinstance(obj, note.NoteOwner):
                for theNote in obj.notes():
                    mapCategorizables(theNote, resultMap, categoryMap)
            if isinstance(obj, attachment.AttachmentOwner):
                for theAttachment in obj.attachments():
                    mapCategorizables(theAttachment, resultMap, categoryMap)

        categorizableMap = dict()
        categoryMap = dict()
        for theCategory in categories:
            mapCategorizables(theCategory, categorizableMap, categoryMap)
        for theTask in tasks:
            mapCategorizables(theTask, categorizableMap, categoryMap)
        for theNote in notes:
            mapCategorizables(theNote, categorizableMap, categoryMap)

        event = patterns.Event()
        for categoryId, categorizableIds in self.__categorizables.items():
            theCategory = categoryMap[categoryId]
            for categorizableId in categorizableIds:
                if categorizableMap.has_key(categorizableId):
                    theCategorizable = categorizableMap[categorizableId]
                    theCategory.addCategorizable(theCategorizable)
                    theCategorizable.addCategory(theCategory, event=event)
        event.send()

    def __parse_category_nodes(self, node):
        return [self.__parse_category_node(child) \
                for child in node.findall('category')]
        
    def __parse_note_nodes(self, node):
        return [self.__parse_note_node(child) for child in node.findall('note')]

    def __parse_category_node(self, category_node):
        ''' Recursively parse the categories from the node and return a 
            category instance. '''
        kwargs = self.__parse_base_composite_attributes(category_node, 
            self.__parse_category_nodes)
        notes = self.__parse_note_nodes(category_node)
        filtered = self.__parse_boolean(category_node.attrib.get('filtered', 
                                                                 'False'))
        exclusive = self.__parse_boolean(category_node.attrib.get(\
                                         'exclusiveSubcategories', 'False'))
        kwargs.update(dict(notes=notes, filtered=filtered,
                           exclusiveSubcategories=exclusive))
        if self.__tskversion < 19:
            categorizable_ids = category_node.attrib.get('tasks', '')
        else:
            categorizable_ids = category_node.attrib.get('categorizables', '')
        if self.__tskversion > 20:
            kwargs['attachments'] = self.__parse_attachments(category_node)
        theCategory = category.Category(**kwargs) # pylint: disable=W0142
        self.__categorizables.setdefault(theCategory.id(), list()).extend(categorizable_ids.split(' '))
        return self.__save_modification_datetime(theCategory)
                      
    def __parse_category_nodes_from_task_nodes(self, root):
        ''' In tskversion <=13 category nodes were subnodes of task nodes. '''
        task_nodes = root.findall('.//task')
        category_mapping = \
            self.__parse_category_nodes_within_task_nodes(task_nodes)
        subject_category_mapping = {}
        for task_id, categories in category_mapping.items():
            for subject in categories:
                if subject in subject_category_mapping:
                    cat = subject_category_mapping[subject]
                else:
                    cat = category.Category(subject)
                    subject_category_mapping[subject] = cat
                self.__categorizables.setdefault(cat.id(), list()).append(task_id)
        return subject_category_mapping.values()
    
    def __parse_category_nodes_within_task_nodes(self, task_nodes):
        ''' In tskversion <=13 category nodes were subnodes of task nodes. '''
        category_mapping = {}
        for node in task_nodes:
            task_id = node.attrib['id']
            categories = [child.text for child in node.findall('category')]
            category_mapping.setdefault(task_id, []).extend(categories)
        return category_mapping
        
    def _parse_task_node(self, task_node):
        '''Recursively parse the node and return a task instance. '''
            
        planned_start_datetime_attribute_name = 'startdate' if self.tskversion() <= 33 else 'plannedstartdate'
        kwargs = self.__parse_base_composite_attributes(task_node,
                                                        self.__parse_task_nodes)
        kwargs.update(dict(
            plannedStartDateTime=date.parseDateTime(task_node.attrib.get(planned_start_datetime_attribute_name, ''), 
                                                    *self.defaultStartTime),
            dueDateTime=parseAndAdjustDateTime(task_node.attrib.get('duedate', ''), 
                                           *self.defaultEndTime),
            actualStartDateTime=date.parseDateTime(task_node.attrib.get('actualstartdate', ''),
                                                   *self.defaultStartTime),
            completionDateTime=date.parseDateTime(task_node.attrib.get('completiondate', ''), 
                                                  *self.defaultEndTime),
            percentageComplete=self.__parse_int_attribute(task_node, 
                                                       'percentageComplete'),
            budget=date.parseTimeDelta(task_node.attrib.get('budget', '')),
            priority=self.__parse_int_attribute(task_node, 'priority'),
            hourlyFee=float(task_node.attrib.get('hourlyFee', '0')),
            fixedFee=float(task_node.attrib.get('fixedFee', '0')),
            reminder=self.__parse_datetime(task_node.attrib.get('reminder', '')),
            reminderBeforeSnooze=self.__parse_datetime(task_node.attrib.get('reminderBeforeSnooze', '')),
            # Ignore prerequisites for now, they'll be resolved later
            prerequisites=[],
            shouldMarkCompletedWhenAllChildrenCompleted=self.__parse_boolean(task_node.attrib.get('shouldMarkCompletedWhenAllChildrenCompleted', '')),
            efforts=self.__parse_effort_nodes(task_node),
            notes=self.__parse_note_nodes(task_node),
            recurrence=self.__parse_recurrence(task_node)))
        self.__prerequisites[kwargs['id']] = [id_ for id_ in task_node.attrib.get('prerequisites', '').split(' ') if id_]
        if self.__tskversion > 20:
            kwargs['attachments'] = self.__parse_attachments(task_node)
        return self.__save_modification_datetime(task.Task(**kwargs))  # pylint: disable=W0142
        
    def __parse_recurrence(self, task_node):
        ''' Parse the recurrence from the node and return a recurrence 
            instance. '''
        if self.__tskversion <= 19:
            parse_kwargs = self.__parse_recurrence_attributes_from_task_node
        else:
            parse_kwargs = self.__parse_recurrence_node
        return date.Recurrence(**parse_kwargs(task_node))
    
    def __parse_recurrence_node(self, task_node):
        ''' Since tskversion >= 20, recurrence information is stored in a 
            separate node. '''
        kwargs = dict(unit='', amount=1, count=0, maximum=0, stop_datetime=None,
                      sameWeekday=False)
        node = task_node.find('recurrence')
        if node is not None:
            kwargs = dict(unit=node.attrib.get('unit', ''),
                amount=int(node.attrib.get('amount', '1')),
                count=int(node.attrib.get('count', '0')),
                maximum=int(node.attrib.get('max', '0')),
                stop_datetime=self.__parse_datetime(node.attrib.get('stop_datetime', '')),
                sameWeekday=self.__parse_boolean(node.attrib.get('sameWeekday', 
                                                                 'False')),
                recurBasedOnCompletion=self.__parse_boolean(node.attrib.get('recurBasedOnCompletion', 'False')))
        return kwargs
    
    @staticmethod                           
    def __parse_recurrence_attributes_from_task_node(task_node):
        ''' In tskversion <= 19 recurrence information was stored as attributes
            of task nodes. '''
        return dict(unit=task_node.attrib.get('recurrence', ''),
            count=int(task_node.attrib.get('recurrenceCount', '0')),
            amount=int(task_node.attrib.get('recurrenceFrequency', '1')),
            maximum=int(task_node.attrib.get('maxRecurrenceCount', '0')))
    
    def __parse_note_node(self, note_node):
        ''' Parse the attributes and child notes from the noteNode. '''
        kwargs = self.__parse_base_composite_attributes(note_node, 
                                                    self.__parse_note_nodes)
        if self.__tskversion > 20:
            kwargs['attachments'] = self.__parse_attachments(note_node)
        return self.__save_modification_datetime(note.Note(**kwargs))  # pylint: disable=W0142
    
    def __parse_base_attributes(self, node):
        ''' Parse the attributes all composite domain objects share, such as
            id, subject, description, and return them as a 
            keyword arguments dictionary that can be passed to the domain 
            object constructor. '''
        bg_color_attribute = 'color' if self.__tskversion <= 27 else 'bgColor'
        attributes = dict(id=node.attrib.get('id', ''),
            creationDateTime=self.__parse_datetime(node.attrib.get('creationDateTime', 
                                                                   '1-1-1 0:0')),
            modificationDateTime=self.__parse_datetime(node.attrib.get('modificationDateTime',
                                                                       '1-1-1 0:0')),
            subject=node.attrib.get('subject', ''),
            description=self.__parse_description(node),
            fgColor=self.__parse_tuple(node.attrib.get('fgColor', ''), None),
            bgColor=self.__parse_tuple(node.attrib.get(bg_color_attribute, ''), 
                                     None),
            font=self.__parse_font_description(node.attrib.get('font', '')),
            icon=self.__parse_icon(node.attrib.get('icon', '')),
            selectedIcon=self.__parse_icon(node.attrib.get('selectedIcon', '')),
            ordering=long(node.attrib.get('ordering', '0L')))

        if self.__tskversion <= 20:
            attributes['attachments'] = \
                self.__parse_attachments_before_version21(node)
        if self.__tskversion >= 22:
            attributes['status'] = int(node.attrib.get('status', '1'))

        return attributes
    
    def __parse_base_composite_attributes(self, node, parse_children, 
                                          *parse_children_args):
        ''' Same as __parse_base_attributes, but also parse children and 
            expandedContexts. '''
        kwargs = self.__parse_base_attributes(node)
        kwargs['children'] = parse_children(node, *parse_children_args)
        expanded_contexts = node.attrib.get('expandedContexts', '')
        kwargs['expandedContexts'] = self.__parse_tuple(expanded_contexts, [])
        return kwargs

    def __parse_attachments_before_version21(self, parent):
        ''' Parse the attachments from the node and return the attachment 
            instances. '''
        path, name = os.path.split(os.path.abspath(self.__fd.name))  # pylint: disable=E1103
        name = os.path.splitext(name)[0]
        attdir = os.path.normpath(os.path.join(path, name + '_attachments'))

        attachments = []
        for node in parent.findall('attachment'):
            if self.__tskversion <= 16:
                args = (node.text,)
                kwargs = dict()
            else:
                args = (os.path.join(attdir, node.find('data').text), 
                        node.attrib['type'])
                description = self.__parse_description(node)
                kwargs = dict(subject=description,
                              description=description)
            try:
                # pylint: disable=W0142
                attachments.append(attachment.AttachmentFactory(*args, 
                                                                **kwargs))  
            except IOError:
                # Mail attachment, file doesn't exist. Ignore this.
                pass
        return attachments

    def __parse_effort_nodes(self, node):
        ''' Parse all effort records from the node. '''
        return [self.__parse_effort_node(effort_node) \
                for effort_node in node.findall('effort')]

    def __parse_effort_node(self, node):
        ''' Parse an effort record from the node. '''
        kwargs = {}
        if self.__tskversion >= 22:
            kwargs['status'] = int(node.attrib.get('status', '1'))
        if self.__tskversion >= 29:
            kwargs['id'] = node.attrib['id']
        start = node.attrib.get('start', '')
        stop = node.attrib.get('stop', '')
        description = self.__parse_description(node)
        # task=None because it is set when the effort is actually added to the
        # task by the task itself. This way no events are sent for changing the
        # effort owner, which is good.
        # pylint: disable=W0142 
        return effort.Effort(task=None, start=date.parseDateTime(start),
            stop=date.parseDateTime(stop), description=description, **kwargs)

    def __parse_syncml_node(self, nodes, guid):
        ''' Parse the SyncML node from the nodes. '''
        syncml_config = createDefaultSyncConfig(guid)

        node_name = 'syncmlconfig'
        if self.__tskversion < 25:
            node_name = 'syncml'

        for node in nodes.findall(node_name):
            self.__parse_syncml_nodes(node, syncml_config)
        return syncml_config

    def __parse_syncml_nodes(self, node, config_node):
        ''' Parse the SyncML nodes from the node. '''
        for child_node in node:
            if child_node.tag == 'property':
                config_node.set(child_node.attrib['name'], 
                                self.__parse_text(child_node))
            else:
                for child_config_node in config_node.children():
                    if child_config_node.name == child_node.tag:
                        break
                else:
                    tag = child_node.tag
                    child_config_node = SyncMLConfigNode(tag)
                    config_node.addChild(child_config_node)
                self.__parse_syncml_nodes(child_node, child_config_node) 

    def __parse_guid_node(self, node):
        ''' Parse the GUID from the node. '''
        guid = self.__parse_text(node).strip()
        return guid if guid else generate()
        
    def __parse_attachments(self, node):
        ''' Parse the attachments from the node. '''
        attachments = []
        for child_node in node.findall('attachment'):
            try:
                attachments.append(self.__parse_attachment(child_node))
            except IOError:
                pass
        return attachments

    def __parse_attachment(self, node):
        ''' Parse the attachment from the node. '''
        kwargs = self.__parse_base_attributes(node)
        kwargs['notes'] = self.__parse_note_nodes(node)

        if self.__tskversion <= 22:
            path, name = os.path.split(os.path.abspath(\
                                       self.__fd.name))  # pylint: disable=E1103
            name, ext = os.path.splitext(name)
            attdir = os.path.normpath(os.path.join(path, name + '_attachments'))
            location = os.path.join(attdir, node.attrib['location'])
        else:
            if 'location' in node.attrib:
                location = node.attrib['location']
            else:
                data_node = node.find('data')

                if data_node is None:
                    raise ValueError('Neither location or data are defined '
                                     'for this attachment.')

                data = self.__parse_text(data_node)
                ext = data_node.attrib['extension']

                location = sessiontempfile.get_temp_file(suffix=ext)
                file(location, 'wb').write(data.decode('base64'))

                if os.name == 'nt':
                    os.chmod(location, stat.S_IREAD)

        return self.__save_modification_datetime(attachment.AttachmentFactory(location,  # pylint: disable=W0142
                                                                              node.attrib['type'], **kwargs))

    def __parse_description(self, node):
        ''' Parse the description from the node. '''
        if self.__tskversion <= 6:
            description = node.attrib.get('description', '')
        else:
            description = self.__parse_text(node.find('description'))
        return description
    
    def __parse_text(self, node):
        ''' Parse the text from a node. '''
        text = u'' if node is None else node.text or u''
        if self.__tskversion >= 24:
            # Strip newlines
            if text.startswith('\n'):
                text = text[1:]
            if text.endswith('\n'):
                text = text[:-1]
        return text
                    
    @classmethod
    def __parse_int_attribute(cls, node, attribute_name, default_value=0):
        ''' Parse the integer attribute with the specified name from the 
            node. In case of failure, return the default value. '''
        text = node.attrib.get(attribute_name, '0')
        return cls.__parse(text, int, default_value)
        
    @classmethod
    def __parse_datetime(cls, text):
        ''' Parse a datetime from the text. '''
        return cls.__parse(text, date.parseDateTime, None)
    
    def __parse_font_description(self, text, default_value=None):
        ''' Parse a font from the text. In case of failure, return the default
            value. '''
        if text:
            font = wx.FontFromNativeInfoString(text)
            if font and font.IsOk():
                if font.GetPointSize() < 4:
                    font.SetPointSize(self.__default_font_size)
                return font
        return default_value
    
    @staticmethod
    def __parse_icon(text):
        ''' Parse an icon name from the text. '''
        # Parse is a big word, we just need to fix one particular icon
        return 'clock_alarm_icon' if text == 'clock_alarm' else text
    
    @classmethod
    def __parse_boolean(cls, text, default_value=None):
        ''' Parse a boolean from the text. In case of failure, return the
            default value. '''
        def text_to_boolean(text):
            ''' Transform 'True' to True and 'False' to False, raise a 
                ValueError for any other text. '''
            if text in ('True', 'False'):
                return text == 'True'
            else:
                raise ValueError("Expected 'True' or 'False', got '%s'" % text)
        return cls.__parse(text, text_to_boolean, default_value)
        
    @classmethod
    def __parse_tuple(cls, text, default_value=None):
        ''' Parse a tuple from the text. In case of failure, return the default
            value. '''
        if text.startswith('(') and text.endswith(')'):
            return cls.__parse(text, eval, default_value)
        else:
            return default_value
    
    @staticmethod
    def __parse(text, parse_function, default_value):
        ''' Parse the text using the parse function. In case of failure, return
            the default value. '''
        try:
            return parse_function(text)
        except ValueError:
            return default_value
        
    def __save_modification_datetime(self, item):
        ''' Save the modification date time of the item for later restore. '''
        self.__modification_datetimes[item] = item.modificationDateTime()
        return item


class ChangesXMLReader(object):
    def __init__(self, fd):
        self.__fd = fd

    def read(self):
        allChanges = dict()
        tree = ET.parse(self.__fd)
        for devNode in tree.getroot().findall('device'):
            id_ = devNode.attrib['guid']
            mon = ChangeMonitor(id_)
            for objNode in devNode.findall('obj'):
                if objNode.text:
                    changes = set(objNode.text.split(','))
                else:
                    changes = set()
                mon.setChanges(objNode.attrib['id'], changes)
            allChanges[id_] = mon
        return allChanges


class TemplateXMLReader(XMLReader):
    def read(self):
        return super(TemplateXMLReader, self).read()[0][0]

    def _parse_task_node(self, task_node):
        attrs = dict()
        attribute_renames = dict(startdate='plannedstartdate')
        for name in ['startdate', 'plannedstartdate', 'duedate', 
                     'completiondate', 'reminder']:
            new_name = attribute_renames.get(name, name)
            template_name = name + 'tmpl'
            if template_name in task_node.attrib:
                if self.tskversion() < 32:
                    value = TemplateXMLReader.convert_old_format(task_node.attrib[template_name])
                else:
                    value = task_node.attrib[template_name]
                attrs[new_name] = value
                task_node.attrib[new_name] = str(nlTimeExpression.parseString(value).calculatedTime)
            elif new_name not in attrs:
                attrs[new_name] = None
        if 'subject' in task_node.attrib:
            task_node.attrib['subject'] = translate(task_node.attrib['subject'])
        parsed_task = super(TemplateXMLReader, self)._parse_task_node(task_node)
        for name, value in attrs.items():
            setattr(parsed_task, name + 'tmpl', value)
        return parsed_task

    @staticmethod
    def convert_old_format(expr, now=date.Now):
        # Built-in templates:
        built_in_templates = {'Now()': 'now', 
                              'Now().endOfDay()': '11:59 PM today',
                              'Now().endOfDay() + oneDay': '11:59 PM tomorrow',
                              'Today()': '00:00 AM today',
                              'Tomorrow()': '11:59 PM tomorrow'}
        if expr in built_in_templates:
            return built_in_templates[expr]
        # Not a built in template:
        new_datetime = eval(expr, date.__dict__)
        if isinstance(new_datetime, date.date.RealDate):
            new_datetime = date.DateTime(new_datetime.year, new_datetime.month, 
                                        new_datetime.day)
        delta = new_datetime - now()
        minutes = delta.minutes()
        if minutes < 0:
            return '%d minutes ago' % (-minutes)
        else:
            return '%d minutes from now' % minutes



#!/usr/bin/env python

# GUID.py
# Version 2.6
#
# Copyright (c) 2006 Conan C. Albrecht
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is furnished
# to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
# FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.



##################################################################################################
###   A globally-unique identifier made up of time and ip and 8 digits for a counter:
###   each GUID is 40 characters wide
###
###   A globally unique identifier that combines ip, time, and a counter.  Since the
###   time is listed first, you can sort records by guid.  You can also extract the time
###   and ip if needed.
###
###   Since the counter has eight hex characters, you can create up to
###   0xffffffff (4294967295) GUIDs every millisecond.  If your processor
###   is somehow fast enough to create more than that in a millisecond (looking
###   toward the future, of course), the function will wait until the next
###   millisecond to return.
###
###   GUIDs make wonderful database keys.  They require no access to the
###   database (to get the max index number), they are extremely unique, and they sort
###   automatically by time.   GUIDs prevent key clashes when merging
###   two databases together, combining data, or generating keys in distributed
###   systems.
###
###   There is an Internet Draft for UUIDs, but this module does not implement it.
###   If the draft catches on, perhaps I'll conform the module to it.
###


# Changelog
# Sometime, 1997     Created the Java version of GUID
#                    Went through many versions in Java
# Sometime, 2002     Created the Python version of GUID, mirroring the Java version
# November 24, 2003  Changed Python version to be more pythonic, took out object and made just a module
# December 2, 2003   Fixed duplicating GUIDs.  Sometimes they duplicate if multiples are created
#                    in the same millisecond (it checks the last 100 GUIDs now and has a larger random part)
# December 9, 2003   Fixed MAX_RANDOM, which was going over sys.maxint
# June 12, 2004      Allowed a custom IP address to be sent in rather than always using the
#                    local IP address.
# November 4, 2005   Changed the random part to a counter variable.  Now GUIDs are totally
#                    unique and more efficient, as long as they are created by only
#                    on runtime on a given machine.  The counter part is after the time
#                    part so it sorts correctly.
# November 8, 2005   The counter variable now starts at a random long now and cycles
#                    around.  This is in case two guids are created on the same
#                    machine at the same millisecond (by different processes).  Even though
#                    it is possible the GUID can be created, this makes it highly unlikely
#                    since the counter will likely be different.
# November 11, 2005  Fixed a bug in the new IP getting algorithm.  Also, use IPv6 range
#                    for IP when we make it up (when it's no accessible)
# November 21, 2005  Added better IP-finding code.  It finds IP address better now.
# January 5, 2006    Fixed a small bug caused in old versions of python (random module use)

import math
import socket
import random
import sys
import time
import threading



#############################
###   global module variables

#Makes a hex IP from a decimal dot-separated ip (eg: 127.0.0.1)
make_hexip = lambda ip: ''.join(["%04x" % long(i) for i in ip.split('.')]) # leave space for ip v6 (65K in each sub)

MAX_COUNTER = 0xfffffffe
counter = 0L
firstcounter = MAX_COUNTER
lasttime = 0
ip = ''
lock = threading.RLock()
try:  # only need to get the IP addresss once
    ip = socket.getaddrinfo(socket.gethostname(),0)[-1][-1][0]
    hexip = make_hexip(ip)
except: # if we don't have an ip, default to someting in the 10.x.x.x private range
    ip = '10'
    rand = random.Random()
    for i in range(3):
        ip += '.' + str(rand.randrange(1, 0xffff))  # might as well use IPv6 range if we're making it up
    hexip = make_hexip(ip)


#################################
###   Public module functions

def generate(ip=None):
    '''Generates a new guid.  A guid is unique in space and time because it combines
       the machine IP with the current time in milliseconds.  Be careful about sending in
       a specified IP address because the ip makes it unique in space.  You could send in
       the same IP address that is created on another machine.
    '''
    global counter, firstcounter, lasttime
    lock.acquire() # can't generate two guids at the same time
    try:
        parts = []

        # do we need to wait for the next millisecond (are we out of counters?)
        now = long(time.time() * 1000)
        while lasttime == now and counter == firstcounter:
            time.sleep(.01)
            now = long(time.time() * 1000)

        # time part
        parts.append("%016x" % now)

        # counter part
        if lasttime != now:  # time to start counter over since we have a different millisecond
            firstcounter = long(random.uniform(1, MAX_COUNTER))  # start at random position
            counter = firstcounter
        counter += 1
        if counter > MAX_COUNTER:
            counter = 0
        lasttime = now
        parts.append("%08x" % (counter))

        # ip part
        parts.append(hexip)

        # put them all together
        return ''.join(parts)
    finally:
        lock.release()


def extract_time(guid):
    '''Extracts the time portion out of the guid and returns the
       number of seconds since the epoch as a float'''
    return float(long(guid[0:16], 16)) / 1000.0


def extract_counter(guid):
    '''Extracts the counter from the guid (returns the bits in decimal)'''
    return int(guid[16:24], 16)


def extract_ip(guid):
    '''Extracts the ip portion out of the guid and returns it
       as a string like 10.10.10.10'''
    # there's probably a more elegant way to do this
    thisip = []
    for index in range(24, 40, 4):
        thisip.append(str(int(guid[index: index + 4], 16)))
    return '.'.join(thisip)



### TESTING OF GUID CLASS ###
if __name__ == "__main__":
    guids = []
    for i in range(10):  # calculate very fast so people can see the counter in action
        guid = generate()
        guids.append(guid)
    for guid in guids:
        print "GUID:", guid
        guidtime = extract_time(guid)
        print "\tTime:   ", time.strftime('%a, %d %b %Y %H:%M:%S', time.localtime(guidtime)), '(millis: ' + str(round(guidtime - long(guidtime), 3)) + ')'
        print "\tIP:     ", extract_ip(guid)
        print "\tCounter:", extract_counter(guid)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import os, stat, codecs
from taskcoachlib import persistence


if os.name == 'nt':
    from win32com.client import GetActiveObject # pylint: disable=F0401

    def getCurrentSelection():
        selection = GetActiveObject('Outlook.Application').ActiveExplorer().Selection
        filenames = []
        for n in xrange(1, selection.Count + 1):
            filename = persistence.get_temp_file(suffix='.eml')
            saveItem(selection.Item(n), filename)
            filenames.append(filename)
        return filenames

    def saveItem(item, filename):
        body = item.Body
        encoding = 'iso-8859-1'
        try:
            codecs.encode(body, encoding)
        except UnicodeEncodeError:
            encoding = 'utf-8'
        mailFile = codecs.open(filename, 'wb', encoding)
        try:
            mailFile.write(emailHeaders(item, encoding) + body)
        finally:
            mailFile.close()
            os.chmod(filename, stat.S_IREAD)

    def emailHeaders(item, encoding, lineSep=u'\r\n'):
        headers = []
        headers.append(u'subject: %s'%item.Subject)
        headers.append(u'X-Outlook-ID: %s'%item.EntryID)
        headers.append(u'Content-Transfer-Encoding: %s'%encoding)
        headers.append(lineSep)
        return lineSep.join(headers)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test, wx, time
from taskcoachlib import gui, config, persistence
from taskcoachlib.domain import task, date, effort


class ReminderControllerUnderTest(gui.ReminderController):
    def __init__(self, *args, **kwargs):
        self.messages = []
        self.userAttentionRequested = False
        super(ReminderControllerUnderTest, self).__init__(*args, **kwargs)
        
    def showReminderMessage(self, message): # pylint: disable=W0221
        class DummyDialog(object):
            def __init__(self, *args, **kwargs):
                pass
            def Bind(self, *args, **kwargs):
                pass
            def Show(self):
                pass
        super(ReminderControllerUnderTest, self).showReminderMessage(message, DummyDialog)
        self.messages.append(message)
    
    def requestUserAttention(self):
        self.userAttentionRequested = True

        
class DummyWindow(wx.Frame):
    def __init__(self):
        super(DummyWindow, self).__init__(None)
        self.taskFile = persistence.TaskFile()
    

class ReminderControllerTestCase(test.TestCase):
    def setUp(self):
        task.Task.settings = settings = config.Settings(load=False)
        self.taskList = task.TaskList()
        self.effortList = effort.EffortList(self.taskList)
        self.dummyWindow = DummyWindow()
        self.reminderController = ReminderControllerUnderTest(self.dummyWindow, 
            self.taskList, self.effortList, settings)
        self.nowDateTime = date.DateTime.now()
        self.reminderDateTime = self.nowDateTime + date.ONE_HOUR
        
    def tearDown(self):
        super(ReminderControllerTestCase, self).tearDown()
        self.dummyWindow.taskFile.close()
        self.dummyWindow.taskFile.stop()
        

class ReminderControllerTest(ReminderControllerTestCase):
    def setUp(self):
        super(ReminderControllerTest, self).setUp()
        self.task = task.Task('Task')
        self.taskList.append(self.task)
        
    def testSetTaskReminderSchedulesJob(self):
        self.task.setReminder(self.reminderDateTime)
        self.failUnless(date.Scheduler().get_jobs())
        
    @test.skipOnTwistedVersions('12.')
    def testAfterReminderJobIsRemovedFromScheduler(self):
        self.task.setReminder(date.Now() + date.TimeDelta(seconds=1))
        self.failUnless(date.Scheduler().get_jobs())
        t0 = time.time()
        from twisted.internet import reactor
        while time.time() - t0 < 1.1:
            reactor.iterate()
        self.failIf(date.Scheduler().get_jobs())
        
    def testAddTaskWithReminderSchedulesJob(self):
        taskWithReminder = task.Task('Task with reminder', 
                                     reminder=self.reminderDateTime)
        self.taskList.append(taskWithReminder)
        self.failUnless(date.Scheduler().get_jobs())
                
    def testRemoveTaskWithReminderRemovesClockEventFromPublisher(self):
        self.task.setReminder(self.reminderDateTime)
        job = date.Scheduler().get_jobs()[0]
        self.taskList.remove(self.task)
        self.failIf(job in date.Scheduler().get_jobs())
                
    def testChangeReminderRemovesOldReminder(self):
        self.task.setReminder(self.reminderDateTime)
        job = date.Scheduler().get_jobs()[0]
        self.task.setReminder(self.reminderDateTime + date.ONE_HOUR)
        jobs = date.Scheduler().get_jobs()
        self.assertEqual(len(jobs), 1)
        self.failIf(job is jobs[0])
        
    def testMarkTaskCompletedRemovesReminder(self):
        self.task.setReminder(self.reminderDateTime)
        self.failUnless(date.Scheduler().get_jobs())
        self.task.setCompletionDateTime(date.Now())
        self.failIf(date.Scheduler().get_jobs())
        
    def dummyCloseEvent(self, snoozeTimeDelta=None, openAfterClose=False):
        class DummySnoozeOptions(object):
            Selection = 0
            def GetClientData(self, *args): # pylint: disable=W0613
                return snoozeTimeDelta
        class DummyDialog(object):
            task = self.task
            openTaskAfterClose = openAfterClose
            ignoreSnoozeOption = False
            snoozeOptions = DummySnoozeOptions()
            def Destroy(self):
                pass
        class DummyEvent(object):
            EventObject = DummyDialog()
            def Skip(self):
                pass
        return DummyEvent()
    
    def testOnCloseReminderResetsReminder(self):
        self.task.setReminder(self.reminderDateTime)
        self.reminderController.onCloseReminderDialog(self.dummyCloseEvent(), 
                                                     show=False)
        self.assertEqual(None, self.task.reminder())

    def testOnCloseReminderSetsReminder(self):
        self.task.setReminder(self.reminderDateTime)
        self.reminderController.onCloseReminderDialog(\
            self.dummyCloseEvent(date.ONE_HOUR), show=False)
        self.failUnless(abs(self.nowDateTime + date.ONE_HOUR - self.task.reminder()) \
                        < date.TimeDelta(seconds=5))

    def testOnCloseMayOpenTask(self):
        self.task.setReminder(self.reminderDateTime)
        frame = self.reminderController.onCloseReminderDialog(\
            self.dummyCloseEvent(openAfterClose=True), show=False)
        self.failUnless(frame)
        
    def testOnWakeDoesNotRequestUserAttentionWhenThereAreNoReminders(self):
        self.reminderController.onReminder()
        self.failIf(self.reminderController.userAttentionRequested)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import application
from taskcoachlib.domain import task


class MockWxApp(object):
    def SetAppName(self, *args, **kwargs):
        pass

    def SetVendorName(self, *args, **kwargs):
        pass

    def ProcessIdle(self):
        pass
    
    def ExitMainLoop(self):
        pass
    
    def SetTopWindow(self, *args):
        pass

    def onQuit(self):
        pass
    
    
class App(application.Application):
    def __init__(self, args=None): # pylint: disable=W0231
        self._options = None
        self._args = args or []
        self._Application__wx_app = MockWxApp()
        self.init()
        self.sessionMonitor = None
        application.Application.instance = self

    def initTwisted(self):
        pass

    def stopTwisted(self):
        pass

    def registerApp(self):
        pass

    def init(self): # pylint: disable=W0221
        super(App, self).init(loadSettings=False, loadTaskFile=False)

    def addTask(self):
        self.task = task.Task('Task')
        self.taskFile.tasks().extend([self.task])

    def addTasks(self):
        # pylint: disable=W0201
        self.parent = task.Task('Parent')
        self.child = task.Task('Child')
        self.parent.addChild(self.child)
        self.taskFile.tasks().extend([self.parent])





'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import taskcoachlib.thirdparty.aui as aui
import wx


class GridCursor:
    ''' Utility class to help when adding controls to a GridBagSizer. '''
    
    def __init__(self, columns):
        self.__columns = columns
        self.__nextPosition = (0, 0)
    
    def __updatePosition(self, colspan):
        ''' Update the position of the cursor, taking colspan into account. '''
        row, column = self.__nextPosition
        if column == self.__columns - colspan:
            row += 1
            column = 0
        else:
            column += colspan
        self.__nextPosition = (row, column)
                    
    def next(self, colspan=1):
        row, column = self.__nextPosition
        self.__updatePosition(colspan)
        return row, column

    def maxRow(self):
        row, column = self.__nextPosition
        return max(0, row - 1) if column == 0 else row


class BookPage(wx.Panel):
    ''' A page in a notebook. '''
    def __init__(self, parent, columns, growableColumn=None, *args, **kwargs):
        super(BookPage, self).__init__(parent, style=wx.TAB_TRAVERSAL, 
            *args, **kwargs)
        self._sizer = wx.GridBagSizer(vgap=5, hgap=5)
        self._columns = columns
        self._position = GridCursor(columns)
        if growableColumn is None:
            self._growableColumn = columns - 1
        else:
            self._growableColumn = growableColumn
        self._borderWidth = 5

    def fit(self):
        self.SetSizerAndFit(self._sizer)

    def __defaultFlags(self, controls):
        ''' Return the default flags for placing a list of controls. '''
        labelInFirstColumn = type(controls[0]) in [type(''), type(u'')]
        flags = []
        for columnIndex in range(len(controls)):
            flag = wx.ALL | wx.ALIGN_CENTER_VERTICAL
            if columnIndex == 0 and labelInFirstColumn:
                flag |= wx.ALIGN_LEFT
            else:
                flag |= wx.ALIGN_RIGHT | wx.EXPAND
            flags.append(flag)
        return flags

    def __determineFlags(self, controls, flagsPassed):
        ''' Return a merged list of flags by overriding the default
            flags with flags passed by the caller. '''
        flagsPassed = flagsPassed or [None] * len(controls)
        defaultFlags = self.__defaultFlags(controls)
        return [defaultFlag if flagPassed is None else flagPassed 
                for flagPassed, defaultFlag in zip(flagsPassed, defaultFlags)]
 
    def addEntry(self, *controls, **kwargs):
        ''' Add a number of controls to the page. All controls are
            placed on one row, and together they form one entry. E.g. a
            label, a text field and an explanatory label. The default
            flags for placing the controls can be overridden by
            providing a keyword parameter 'flags'. flags should be a
            list of flags (wx.ALIGN_LEFT and the like). The list may
            contain None for controls that should be placed using the default
            flag. If the flags list is shorter than the number of
            controls it is extended with as much 'None's as needed.
            So, addEntry(aLabel, aTextCtrl, flags=[None, wx.ALIGN_LEFT]) 
            will place the label with the default flag and will place the 
            textCtrl left aligned. '''
        flags = self.__determineFlags(controls, kwargs.get('flags', None))
        controls = [self.__createStaticTextControlIfNeeded(control) \
                    for control in controls if control is not None]
        lastColumnIndex = len(controls) - 1
        for columnIndex, control in enumerate(controls):
            self.__addControl(columnIndex, control, flags[columnIndex], 
                              lastColumn=columnIndex == lastColumnIndex)
            if columnIndex > 0:
                control.MoveAfterInTabOrder(controls[columnIndex - 1])
        if kwargs.get('growable', False):
            self._sizer.AddGrowableRow(self._position.maxRow())
        # Move growable column definition here
        # There are asserts to fail if the column is already
        # marked growable or if there is no column yet created
        if self._growableColumn > -1 and \
               self._growableColumn >= lastColumnIndex:
            self._sizer.AddGrowableCol(self._growableColumn)
            self._growableColumn = -1

            
    def addLine(self):
        line = wx.StaticLine(self)
        self.__addControl(0, line, flag=wx.GROW | wx.ALIGN_CENTER_VERTICAL, 
                          lastColumn=True)

    def __addControl(self, columnIndex, control, flag, lastColumn):
        colspan = max(self._columns - columnIndex, 1) if lastColumn else 1
        self._sizer.Add(control, self._position.next(colspan),
            span=(1, colspan), flag=flag, border=self._borderWidth)
        
    def __createStaticTextControlIfNeeded(self, control):
        if type(control) in [type(''), type(u'')]:
            control = wx.StaticText(self, label=control)
        return control
        

class BookMixin(object):
    ''' Mixin class for *book '''
    
    _bitmapSize = (16, 16)
    pageChangedEvent = 'Subclass responsibility'
    
    def __init__(self, parent, *args, **kwargs):
        super(BookMixin, self).__init__(parent, -1, *args, **kwargs)
        self.Bind(self.pageChangedEvent, self.onPageChanged)
        
    def __getitem__(self, index):
        ''' More pythonic way to get a specific page, also useful for iterating
            over all pages, e.g: for page in notebook: ... '''
        if index < self.GetPageCount():
            return self.GetPage(index)
        else:
            raise IndexError
    
    def onPageChanged(self, event):
        ''' Can be overridden in a subclass to do something useful. '''
        event.Skip()    

    def AddPage(self, page, name, bitmap=None):
        ''' Override AddPage to allow for simply specifying the bitmap name. '''
        bitmap = wx.ArtProvider_GetBitmap(bitmap, wx.ART_MENU, self._bitmapSize)
        super(BookMixin, self).AddPage(page, name, bitmap=bitmap)

    def ok(self, *args, **kwargs):
        for page in self:
            page.ok(*args, **kwargs)
            

class Notebook(BookMixin, aui.AuiNotebook):
    pageChangedEvent = aui.EVT_AUINOTEBOOK_PAGE_CHANGED
    
    def __init__(self, *args, **kwargs):
        defaultStyle = kwargs.get('agwStyle', aui.AUI_NB_DEFAULT_STYLE)
        kwargs['agwStyle'] = defaultStyle & ~aui.AUI_NB_CLOSE_ON_ACTIVE_TAB & \
                             ~aui.AUI_NB_MIDDLE_CLICK_CLOSE 
        super(Notebook, self).__init__(*args, **kwargs)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
import test
from unittests import dummy
from taskcoachlib import widgets

        
class TreeCtrlTestCase(test.wxTestCase):
    onSelect = None
    
    def getFirstTreeItem(self):
        # pylint: disable=E1101
        return self.treeCtrl.GetFirstChild(self.treeCtrl.GetRootItem())[0]
                
    def setUp(self):
        super(TreeCtrlTestCase, self).setUp()
        self.children = dict()
        self.collapsedItems = []
        self.frame.children = lambda item: self.children.get(item, [])
        self.frame.getItemText = lambda item, column: item.subject()
        self.frame.hasColumnImages = lambda column: False
        self.frame.getItemImages = lambda item, column: {wx.TreeItemIcon_Normal: -1}
        self.frame.getIsItemChecked = lambda item: False
        self.frame.getItemExpanded = lambda item: item not in self.collapsedItems
        self.item0 = DummyDomainObject('item 0')
        self.item1 = DummyDomainObject('item 1')
        self.item0_0 = DummyDomainObject('item 0.0')
        self.item0_1 = DummyDomainObject('item 0.1')
        self.item1_0 = DummyDomainObject('item 1.0')


class DummyDomainObject(object):
    def __init__(self, subject):
        self.__subject = subject
        
    def subject(self):
        return self.__subject
    
    # pylint: disable=W0613
    
    def foregroundColor(self, recursive=False):
        return None
    
    def backgroundColor(self, recursive=False):
        return None
    
    def font(self, recursive=False):
        return None
    
    
class CommonTestsMixin(object):
    ''' Tests for all types of trees. '''

    def testCreate(self):
        self.assertEqual(0, len(self.treeCtrl.GetItemChildren()))
    
    def testOneItem(self):
        self.children[None] = [self.item0]
        self.treeCtrl.RefreshAllItems(1)
        self.assertEqual(1, len(self.treeCtrl.GetItemChildren()))
    
    def testTwoItems(self):
        self.children[None] = [self.item0, self.item1]
        self.treeCtrl.RefreshAllItems(2)
        self.assertEqual(2, len(self.treeCtrl.GetItemChildren()))
    
    def testRemoveAllItems(self):
        self.children[None] = [self.item0, self.item1]
        self.treeCtrl.RefreshAllItems(2)
        self.children[None] = []
        self.treeCtrl.RefreshAllItems(0)
        self.assertEqual(0, len(self.treeCtrl.GetItemChildren()))
        
    def testOneParentAndOneChild(self):
        self.children[None] = [self.item0]
        self.children[self.item0] = [self.item0_0]
        self.treeCtrl.RefreshAllItems(2)
        self.assertEqual(1, len(self.treeCtrl.GetItemChildren()))
        self.assertEqual(1, len(self.treeCtrl.GetItemChildren(self.getFirstTreeItem())))
        
    def testOneParentAndTwoChildren(self):
        self.children[None] = [self.item0]
        self.children[self.item0] = [self.item0_0, self.item0_1]
        self.treeCtrl.RefreshAllItems(3)
        self.assertEqual(1, len(self.treeCtrl.GetItemChildren()))
        self.assertEqual(2, len(self.treeCtrl.GetItemChildren(self.getFirstTreeItem())))

    def testAddOneChild(self):
        self.children[None] = [self.item0]
        self.children[self.item0] = [self.item0_0]
        self.treeCtrl.RefreshAllItems(2)
        self.children[self.item0] = [self.item0_0, self.item0_1]
        self.treeCtrl.RefreshAllItems(3)
        self.assertEqual(1, len(self.treeCtrl.GetItemChildren()))
        self.assertEqual(2, len(self.treeCtrl.GetItemChildren(self.getFirstTreeItem())))        
             
    def testDeleteOneChild(self):
        self.children[None] = [self.item0]
        self.children[self.item0] = [self.item0_0, self.item0_1]
        self.treeCtrl.RefreshAllItems(3)
        self.children[self.item0] = [self.item0_0]
        self.treeCtrl.RefreshAllItems(2)
        self.assertEqual(1, len(self.treeCtrl.GetItemChildren()))
        self.assertEqual(1, len(self.treeCtrl.GetItemChildren(self.getFirstTreeItem())))
    
    def testReorderItems(self):
        self.children[None] = [self.item0, self.item1]
        self.treeCtrl.RefreshAllItems(2)
        self.children[None] = [self.item1, self.item0]
        self.treeCtrl.RefreshAllItems(2)
        self.assertEqual('item 1', self.treeCtrl.GetItemText(self.getFirstTreeItem()))
    
    def testReorderChildren(self):
        self.children[None] = [self.item0]
        self.children[self.item0] = [self.item0_0, self.item0_1]
        self.treeCtrl.RefreshAllItems(3)
        self.children[self.item0] = [self.item0_1, self.item0_0]
        self.treeCtrl.RefreshAllItems(3)
        self.assertEqual('item 0.1', self.treeCtrl.GetItemText(self.treeCtrl.GetFirstChild(self.getFirstTreeItem())[0]))
        
    def testReorderParentsAndOneChild(self):
        self.children[None] = [self.item0, self.item1]
        self.children[self.item0] = [self.item0_0]
        self.treeCtrl.RefreshAllItems(3)
        self.children[None] = [self.item1, self.item0]
        self.treeCtrl.RefreshAllItems(3)
        self.assertEqual('item 1', self.treeCtrl.GetItemText(self.getFirstTreeItem()))
    
    def testReorderParentsAndTwoChildren(self):
        self.children[None] = [self.item0, self.item1]
        self.children[self.item0] = [self.item0_0, self.item0_1]
        self.treeCtrl.RefreshAllItems(4)
        self.children[None] = [self.item1, self.item0]
        self.children[self.item0] = [self.item0_1, self.item0_0]
        self.treeCtrl.RefreshAllItems(4)
        self.assertEqual('item 1', self.treeCtrl.GetItemText(self.getFirstTreeItem()))
        self.assertEqual(0, len(self.treeCtrl.GetItemChildren(self.getFirstTreeItem())))
    
    def testRetainSelectionWhenEditingTask(self):
        self.children[None] = [self.item0]
        self.treeCtrl.RefreshAllItems(1)
        item = self.getFirstTreeItem()
        self.treeCtrl.SelectItem(item)
        self.failUnless(self.treeCtrl.IsSelected(item))
        self.children[None] = [self.item0]
        self.treeCtrl.RefreshAllItems(1)
        item = self.getFirstTreeItem()
        self.failUnless(self.treeCtrl.IsSelected(item))
 
    def testRetainSelectionWhenEditingSubTask(self):
        self.children[None] = [self.item0]
        self.children[self.item0] = [self.item0_0]
        self.treeCtrl.RefreshAllItems(2)
        item = self.getFirstTreeItem()
        self.treeCtrl.SelectItem(item)
        self.failUnless(self.treeCtrl.IsSelected(item))
        self.children[self.item0] = [self.item0_0]
        self.treeCtrl.RefreshAllItems(2)
        item = self.getFirstTreeItem()
        self.failUnless(self.treeCtrl.IsSelected(item))
    
    def testRetainSelectionWhenAddingSubTask(self):
        self.children[None] = [self.item0]
        self.treeCtrl.RefreshAllItems(1)
        item = self.getFirstTreeItem()
        self.treeCtrl.SelectItem(item)
        self.failUnless(self.treeCtrl.IsSelected(item))
        self.children[self.item0] = [self.item0_0]
        self.treeCtrl.RefreshAllItems(2)
        item = self.getFirstTreeItem()        
        self.failUnless(self.treeCtrl.IsSelected(item))

    def testRetainSelectionWhenAddingSubTask_TwoToplevelTasks(self):
        self.children[None] = [self.item0, self.item1]
        self.treeCtrl.RefreshAllItems(2)
        item = self.getFirstTreeItem()
        self.treeCtrl.SelectItem(item)
        self.failUnless(self.treeCtrl.IsSelected(item))
        self.children[self.item0] = [self.item0_0]
        self.treeCtrl.RefreshAllItems(3)
        item = self.getFirstTreeItem()        
        self.failUnless(self.treeCtrl.IsSelected(item))
        
    def testRemovingASelectedItemDoesNotMakeAnotherOneSelected(self):
        self.children[None] = [self.item0, self.item1]
        self.treeCtrl.RefreshAllItems(2)
        item = self.getFirstTreeItem()
        self.treeCtrl.SelectItem(item)
        self.failUnless(self.treeCtrl.IsSelected(item))
        self.children[None] = [self.item1]
        self.treeCtrl.RefreshAllItems(1)
        self.failIf(self.treeCtrl.curselection())
        
    def testRefreshItem(self):
        self.children[None] = [self.item0]
        self.treeCtrl.RefreshAllItems(1)
        self.treeCtrl.RefreshItems(self.item0)
        item = self.getFirstTreeItem()
        self.assertEqual('item 0', self.treeCtrl.GetItemText(item))        
    
    def testIsAnyItemCollapsable_NoItems(self):
        self.failIf(self.treeCtrl.isAnyItemCollapsable())
      
    def testIsAnyItemExpandable_NoItems(self):
        self.failIf(self.treeCtrl.isAnyItemExpandable())
        
    def testIsAnyItemCollapsable_OneItem(self):
        self.children[None] = [self.item0]
        self.treeCtrl.RefreshAllItems(1)
        self.failIf(self.treeCtrl.isAnyItemCollapsable())
   
    def testIsAnyItemExpandable_OneItem(self):
        self.children[None] = [self.item0]
        self.treeCtrl.RefreshAllItems(1)
        self.failIf(self.treeCtrl.isAnyItemExpandable())
        
    def testIsAnyItemCollapsable_OneCollapsedParent(self):
        self.children[None] = [self.item0]
        self.children[self.item0] = [self.item0_0]
        self.collapsedItems.append(self.item0)
        self.treeCtrl.RefreshAllItems(2)
        self.failIf(self.treeCtrl.isAnyItemCollapsable())
       
    def testIsAnyItemExpandable_OneCollapsedParent(self):
        self.children[None] = [self.item0]
        self.children[self.item0] = [self.item0_0]
        self.collapsedItems.append(self.item0)
        self.treeCtrl.RefreshAllItems(2)
        self.failUnless(self.treeCtrl.isAnyItemExpandable())
    
    def testIsAnyItemCollapsable_OneExpandedParent(self):
        self.children[None] = [self.item0]
        self.children[self.item0] = [self.item0_0]
        self.treeCtrl.RefreshAllItems(2)
        parent = self.getFirstTreeItem()
        self.treeCtrl.Expand(parent)
        self.failUnless(self.treeCtrl.isAnyItemCollapsable())
           
    def testIsAnyItemExpandable_OneExpandedParent(self):
        self.children[None] = [self.item0]
        self.children[self.item0] = [self.item0_0]
        self.treeCtrl.RefreshAllItems(2)
        parent = self.getFirstTreeItem()
        self.treeCtrl.Expand(parent)
        self.failIf(self.treeCtrl.isAnyItemExpandable())

        
class TreeListCtrlTest(TreeCtrlTestCase, CommonTestsMixin):
    def setUp(self):
        super(TreeListCtrlTest, self).setUp()
        columns = [widgets.Column('subject', 'Subject')]
        self.treeCtrl = widgets.TreeListCtrl(self.frame, columns, self.onSelect, 
            dummy.DummyUICommand(), dummy.DummyUICommand(), dummy.DummyUICommand())
        imageList = wx.ImageList(16, 16)
        for bitmapName in ['led_blue_icon', 'folder_blue_icon']:
            imageList.Add(wx.ArtProvider_GetBitmap(bitmapName, wx.ART_MENU, 
                          (16,16)))
        self.treeCtrl.AssignImageList(imageList) # pylint: disable=E1101
    

class CheckTreeCtrlTest(TreeCtrlTestCase, CommonTestsMixin):
    def setUp(self):
        self.frame.getItemParentHasExclusiveChildren = lambda item: item.subject().startswith('mutual')
        super(CheckTreeCtrlTest, self).setUp()
        columns = [widgets.Column('subject', 'Subject')]
        self.treeCtrl = widgets.CheckTreeCtrl(self.frame, columns,
            self.onSelect, self.onCheck, 
            dummy.DummyUICommand(), dummy.DummyUICommand())
        self.mutual1 = DummyDomainObject('mutual 1')
        self.mutual2 = DummyDomainObject('mutual 2')

    def onCheck(self, event, final):
        pass

    def testCheckParentDoesNotCheckChild(self):
        self.children[None] = [self.item0]
        self.children[self.item0] = [self.item0_0]
        self.treeCtrl.RefreshAllItems(2)
        self.treeCtrl.ExpandAll() # pylint: disable=E1101
        parent = self.getFirstTreeItem()
        self.treeCtrl.CheckItem(parent)
        child = self.treeCtrl.GetItemChildren(parent)[0]
        self.failIf(child.IsChecked())
        
    def testCheckParentOfMutualExclusiveChildrenUnchecksAllChildren(self):
        self.children[None] = [self.item0]
        self.children[self.item0] = [self.mutual1, self.mutual2]
        self.treeCtrl.RefreshAllItems(3)
        self.treeCtrl.ExpandAll() # pylint: disable=E1101
        parent = self.getFirstTreeItem()
        children = self.treeCtrl.GetItemChildren(parent)
        self.treeCtrl.CheckItem(children[0])
        self.treeCtrl.CheckItem(parent)
        for child in children:
            self.failIf(child.IsChecked())
        
    def testCheckParentOfMutualExclusiveChildrenUnchecksAllChildrenRecursively(self):
        self.children[None] = [self.item0]
        self.children[self.item0] = [self.mutual1, self.mutual2]
        self.children[self.mutual1] = [self.item1_0]
        self.treeCtrl.RefreshAllItems(4)
        self.treeCtrl.ExpandAll() # pylint: disable=E1101
        parent = self.getFirstTreeItem()
        children = self.treeCtrl.GetItemChildren(parent, recursively=True)
        grandchild = children[1]
        self.treeCtrl.CheckItem(grandchild)
        self.treeCtrl.CheckItem(parent)
        self.failIf(grandchild.IsChecked())
        
    def testCheckMutualExclusiveChildUnchecksParent(self):
        self.children[None] = [self.item0]
        self.children[self.item0] = [self.mutual1, self.mutual2]
        self.treeCtrl.RefreshAllItems(3)
        self.treeCtrl.ExpandAll() # pylint: disable=E1101
        parent = self.getFirstTreeItem()
        children = self.treeCtrl.GetItemChildren(parent)
        self.treeCtrl.CheckItem(parent)
        self.treeCtrl.CheckItem(children[0])
        self.failIf(self.treeCtrl.IsItemChecked(parent))  # pylint: disable=E1101



'''
Task Coach - Your friendly task manager
Copyright (C) 2011 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''


import os, atexit


class NotifierBase(object):
    def __init__(self):
        super(NotifierBase, self).__init__()

        self._filename = None
        self._path = None
        self._name = None
        self.stamp = None
        atexit.register(self._stop)

    def _stop(self):
        self.stop()

    def stop(self):
        pass # Should be overloaded if needed

    def _check(self, filename):
        return self.stamp is None or \
               (filename and \
                os.path.exists(filename) and \
                os.stat(filename).st_mtime > self.stamp)

    def setFilename(self, filename):
        self._filename = filename
        self.stamp = None
        if filename:
            self._path, self._name = os.path.split(filename)
            if os.path.exists(filename):
                self.stamp = os.stat(filename).st_mtime
        else:
            self._path, self._name = None, None

    def saved(self):
        if self._filename and os.path.exists(self._filename):
            self.stamp = os.stat(self._filename).st_mtime
        else:
            self.stamp = None



#-*- coding: UTF-8

from taskcoachlib.i18n import _

def getDefaultTemplates():
    templates = []
    templates.append(('dueTomorrow', '<?xml version="1.0" ?><?taskcoach release="1.1.0" tskversion="30"?><tasks><task duedatetmpl="Now().endOfDay() + oneDay" startdatetmpl="Now()" status="2" subject="New task due tomorrow"/></tasks>\n'))
    _('New task due tomorrow')
    templates.append(('dueToday', '<?xml version="1.0" ?><?taskcoach release="1.1.0" tskversion="30"?><tasks><task duedatetmpl="Now().endOfDay()" startdatetmpl="Now()" status="2" subject="New task due today"/></tasks>\n'))
    _('New task due today')

    return templates



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from wxSchedulerCore import *


class wxSchedulerPrint( wxSchedulerCore ):

	def __init__( self, dc ):
		super( wxSchedulerPrint, self ).__init__()
		
		self.SetDc( dc )
		
	def Draw( self, page ):
		"""
		Draw object on bitmap
		"""
		if page is None:
			self.DrawBuffer()
		else:
			self.pageNumber = page
			self.DrawBuffer()

		return self._bitmap
		
	def GetSize( self ):
		"""
		Return a wx.Size() object representing the page's size
		"""
		return self.GetDc().GetSize()

	def Refresh( self ):
		pass



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
from wx.lib import hyperlink
from taskcoachlib import meta
from taskcoachlib.i18n import _
from wx.lib import sized_controls


class VersionDialog(sized_controls.SizedDialog):  # pylint: disable=R0904,R0901
    ''' Base class for dialogs that announce a new version (and variants 
        thereof). '''
    title = ''
    
    def __init__(self, *args, **kwargs):
        self.settings = kwargs.pop('settings')
        self.message = kwargs.pop('message')
        version = kwargs.pop('version')
        super(VersionDialog, self).__init__(title=self.title, *args, **kwargs)
        pane = self.GetContentsPane()
        pane.SetSizerType("vertical")
        self.messageInfo = dict(version=version, name=meta.data.name,
                                currentVersion=meta.data.version)
        self.createInterior(pane)
        self.check = wx.CheckBox(pane, label=_('Notify me of new versions.'))
        self.check.SetValue(self.settings.getboolean('version', 'notify'))
        buttonSizer = self.CreateStdDialogButtonSizer(wx.OK)
        self.SetButtonSizer(buttonSizer)
        self.Fit()
        buttonSizer.GetAffirmativeButton().Bind(wx.EVT_BUTTON, self.onClose)
        self.Bind(wx.EVT_CLOSE, self.onClose)
        
    def createInterior(self, pane):
        ''' Create the interior parts of the dialog, i.e. the message for the
            user. '''
        raise NotImplementedError
        
    def onClose(self, event):
        ''' When the user closes the dialog, remember whether (s)he wants to be
            notified of new versions. '''
        event.Skip()
        self.settings.set('version', 'notify', str(self.check.GetValue()))


class NewVersionDialog(VersionDialog):
    title = _('New version of %(name)s available') % dict(name=meta.data.name)
            
    def createInterior(self, panel):
        wx.StaticText(panel, 
            label=_('You are using %(name)s version %(currentVersion)s.') % \
                  self.messageInfo)
        urlPanel = sized_controls.SizedPanel(panel)
        urlPanel.SetSizerType('horizontal')
        wx.StaticText(urlPanel, 
            label=_('Version %(version)s of %(name)s is available from') % \
                  self.messageInfo)
        hyperlink.HyperLinkCtrl(urlPanel, label=meta.data.url)
        
        
class VersionUpToDateDialog(VersionDialog):
    title = _('%(name)s is up to date') % dict(name=meta.data.name)

    def createInterior(self, panel):
        wx.StaticText(panel, 
            label=_('%(name)s is up to date at version %(version)s.') % \
                  self.messageInfo)
        
        
class NoVersionDialog(VersionDialog):
    title = _("Couldn't find out latest version")
        
    def createInterior(self, panel):
        wx.StaticText(panel, label=_("Couldn't find out what the latest "
                      "version of %(name)s is.") % self.messageInfo)
        wx.StaticText(panel, label=self.message)
        
        
class PrereleaseVersionDialog(VersionDialog):
    title = _("Prerelease version")

    def createInterior(self, panel):
        wx.StaticText(panel, label=_('You are using %(name)s prerelease '
                      'version %(currentVersion)s.') % self.messageInfo)
        wx.StaticText(panel, label=_('The latest released version of '
                      '%(name)s is %(version)s.') % self.messageInfo)



import sys

from ..test_backend import BackendBasicTests
from ..py30compat import unittest
from keyring.backends import OS_X

def is_osx_keychain_supported():
    return sys.platform in ('mac','darwin')

@unittest.skipUnless(is_osx_keychain_supported(),
                     "Need OS X")
class OSXKeychainTestCase(BackendBasicTests, unittest.TestCase):

    def init_keyring(self):
        return OS_X.Keyring()

    @unittest.expectedFailure
    def test_delete_present(self):
        """Not implemented"""
        super(OSXKeychainTestCase, self).test_delete_present()

class SecurityCommandTestCase(unittest.TestCase):
    def test_SecurityCommand(self):
        self.assertEqual(OS_X.SecurityCommand('get'), 'get-generic-password')
        self.assertEqual(OS_X.SecurityCommand('set', 'internet'), 'set-internet-password')



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import operating_system
import wx
import textwrap


class ToolTipMixin(object):
    """Subclass this and override OnBeforeShowToolTip to provide
    dynamic tooltip over a control."""

    def __init__(self, *args, **kwargs):
        self.__enabled = kwargs.pop('tooltipsEnabled', True)
        super(ToolTipMixin, self).__init__(*args, **kwargs)

        self.__timer = wx.Timer(self, wx.NewId())

        self.__tip = None
        self.__position = (0, 0)
        self.__text = None
        self.__frozen = True

        self.GetMainWindow().Bind(wx.EVT_MOTION, self.__OnMotion)
        self.GetMainWindow().Bind(wx.EVT_LEAVE_WINDOW, self.__OnLeave)
        self.Bind(wx.EVT_TIMER, self.__OnTimer, id=self.__timer.GetId())        

    def SetToolTipsEnabled(self, enabled):
        self.__enabled = enabled

    def PopupMenu(self, menu):
        self.__frozen = False
        super(ToolTipMixin, self).PopupMenu(menu)
        self.__frozen = True

    def ShowTip(self, x, y):
        # Ensure we're not too big (in the Y direction anyway) for the
        # desktop display area. This doesn't work on Linux because
        # ClientDisplayRect() returns the whole display size, not
        # taking the taskbar into account...

        if self.__frozen:
            theDisplay = wx.Display(wx.Display.GetFromPoint(wx.Point(x, y)))
            displayX, displayY, displayWidth, displayHeight = theDisplay.GetClientArea()
            tipWidth, tipHeight = self.__tip.GetSizeTuple()

            if tipHeight > displayHeight:
                # Too big. Take as much space as possible.
                y = 5
                tipHeight = displayHeight - 10
            elif y + tipHeight > displayY + displayHeight:
                # Adjust y so that the whole tip is visible.
                y = displayY + displayHeight - tipHeight - 5

            if tipWidth > displayWidth:
                x = 5
            elif x + tipWidth > displayX + displayWidth:
                x = displayX + displayWidth - tipWidth - 5

            self.__tip.Show(x, y, tipWidth, tipHeight)

    def DoShowTip(self, x, y, tip):
        self.__tip = tip
        self.ShowTip(x, y)

    def HideTip(self):
        if self.__tip:
            self.__tip.Hide()

    def OnBeforeShowToolTip(self, x, y):
        """Should return a wx.Frame instance that will be displayed as
        the tooltip, or None."""
        raise NotImplementedError  # pragma: no cover

    def __OnMotion(self, event):
        x, y = event.GetPosition()

        self.__timer.Stop()

        if self.__tip is not None:
            self.HideTip()
            self.__tip = None

        if self.__enabled:
            newTip = self.OnBeforeShowToolTip(x, y)
            if newTip is not None:
                self.__tip = newTip
                self.__tip.Bind(wx.EVT_MOTION, self.__OnTipMotion)
                self.__position = (x + 20, y + 10)
                self.__timer.Start(200, True)

        event.Skip()

    def __OnTipMotion(self, event):  # pylint: disable=W0613
        self.HideTip()

    def __OnLeave(self, event):
        self.__timer.Stop()

        if self.__tip is not None:
            self.HideTip()
            self.__tip = None

        event.Skip()

    def __OnTimer(self, event):  # pylint: disable=W0613
        self.ShowTip(*self.GetMainWindow().ClientToScreenXY(*self.__position))


if operating_system.isWindows():
    class ToolTipBase(wx.MiniFrame):
        def __init__(self, parent):
            style = wx.FRAME_NO_TASKBAR | wx.FRAME_FLOAT_ON_PARENT | wx.NO_BORDER
            super(ToolTipBase, self).__init__(parent, wx.ID_ANY, 'Tooltip',
                                              style=style)

        def Show(self, x, y, w, h):  # pylint: disable=W0221
            self.SetDimensions(x, y, w, h)
            super(ToolTipBase, self).Show()

elif operating_system.isMac():
    class ToolTipBase(wx.Frame):
        def __init__(self, parent):  # pylint: disable=E1003
            style = wx.FRAME_NO_TASKBAR | wx.FRAME_FLOAT_ON_PARENT | wx.NO_BORDER
            super(ToolTipBase, self).__init__(parent, wx.ID_ANY, 'ToolTip',
                                              style=style)

            # There are some subtleties on Mac regarding multi-monitor
            # displays...

            self.__maxWidth, self.__maxHeight = 0, 0
            for index in xrange(wx.Display.GetCount()):
                x, y, width, height = wx.Display(index).GetGeometry()
                self.__maxWidth = max(self.__maxWidth, x + width)
                self.__maxHeight = max(self.__maxHeight, y + height)

            self.MoveXY(self.__maxWidth, self.__maxHeight)
            super(ToolTipBase, self).Show()

        def Show(self, x, y, width, height):  # pylint: disable=W0221
            self.SetDimensions(x, y, width, height)

        def Hide(self):  # pylint: disable=W0221
            self.MoveXY(self.__maxWidth, self.__maxHeight)

else:
    class ToolTipBase(wx.PopupWindow):
        def Show(self, x, y, width, height):  # pylint: disable=E1003,W0221
            self.SetDimensions(x, y, width, height)
            super(ToolTipBase, self).Show()


class SimpleToolTip(ToolTipBase):
    def __init__(self, parent):
        super(SimpleToolTip, self).__init__(parent)
        self.data = []
        self.Bind(wx.EVT_PAINT, self.OnPaint)

    def SetData(self, data):
        self.data = self._wrapLongLines(data)
        self.SetSize(self._calculateSize())
        self.Refresh()  # Needed on Mac OS X
        
    def _wrapLongLines(self, data):
        wrappedData = []
        wrapper = textwrap.TextWrapper(width=78)
        for icon, lines in data:
            wrappedLines = []
            for line in lines:
                wrappedLines.extend(wrapper.fill(line).split('\n'))
            wrappedData.append((icon, wrappedLines))
        return wrappedData
        
    def _calculateSize(self):
        dc = wx.ClientDC(self)
        self._setFontBrushAndPen(dc)
        width, height = 0, 0
        for sectionIndex in range(len(self.data)):
            sectionWidth, sectionHeight = self._calculateSectionSize(dc, sectionIndex)
            width = max(width, sectionWidth)
            height += sectionHeight
        return wx.Size(width + 6, height + 6)
    
    def _calculateSectionSize(self, dc, sectionIndex):
        icon, lines = self.data[sectionIndex]
        sectionWidth, sectionHeight = 0, 0
        for line in lines:
            lineWidth, lineHeight = self._calculateLineSize(dc, line)
            sectionHeight += lineHeight + 1
            sectionWidth = max(sectionWidth, lineWidth)
        if 0 < sectionIndex < len(self.data) - 1:
            sectionHeight += 3  # Horizontal space between sections
        if icon:
            sectionWidth += 24  # Reserve width for icon(s)
        return sectionWidth, sectionHeight
    
    def _calculateLineSize(self, dc, line):
        return dc.GetTextExtent(line)

    def OnPaint(self, event):  # pylint: disable=W0613
        dc = wx.PaintDC(self)
        dc.BeginDrawing()
        try:
            self._setFontBrushAndPen(dc)
            self._drawBorder(dc)
            self._drawSections(dc)
        finally:
            dc.EndDrawing()
            
    def _setFontBrushAndPen(self, dc):
        font = wx.SystemSettings_GetFont(wx.SYS_DEFAULT_GUI_FONT)
        textColour = wx.SystemSettings_GetColour(wx.SYS_COLOUR_INFOTEXT)
        backgroundColour = wx.SystemSettings_GetColour(wx.SYS_COLOUR_INFOBK)
        dc.SetFont(font)
        dc.SetTextForeground(textColour)
        dc.SetBrush(wx.Brush(backgroundColour))
        dc.SetPen(wx.Pen(textColour))
        
    def _drawBorder(self, dc):
        width, height = self.GetClientSizeTuple()
        dc.DrawRectangle(0, 0, width, height)
        
    def _drawSections(self, dc):
        y = 3
        for sectionIndex in range(len(self.data)):
            y = self._drawSection(dc, y, sectionIndex)
                    
    def _drawSection(self, dc, y, sectionIndex):
        icon, lines = self.data[sectionIndex]
        if not lines:
            return y
        x = 3
        if sectionIndex != 0:
            y = self._drawSectionSeparator(dc, x, y)
        if icon:
            x = self._drawIcon(dc, icon, x, y)
        topOfSection = y
        bottomOfSection = self._drawTextLines(dc, lines, x, y)
        if icon:
            self._drawIconSeparator(dc, x - 2, topOfSection, bottomOfSection)
        return bottomOfSection
    
    def _drawSectionSeparator(self, dc, x, y):
        y += 1
        width = self.GetClientSizeTuple()[0]
        dc.DrawLine(x, y, width - x, y)
        return y + 2
        
    def _drawIcon(self, dc, icon, x, y):
        bitmap = wx.ArtProvider.GetBitmap(icon, wx.ART_FRAME_ICON, (16, 16))
        dc.DrawBitmap(bitmap, x, y, True)
        return 23  # New x
        
    def _drawTextLines(self, dc, textLines, x, y):
        for textLine in textLines:
            y = self._drawTextLine(dc, textLine, x, y)
        return y
        
    def _drawTextLine(self, dc, textLine, x, y):
        try:
            dc.DrawText(textLine, x, y)
        except:
            raise RuntimeError('Could not draw text %s' % repr(textLine))
        textHeight = dc.GetTextExtent(textLine)[1]
        return y + textHeight + 1
    
    def _drawIconSeparator(self, dc, x, top, bottom):
        ''' Draw a vertical line between the icon and the text. '''
        dc.DrawLine(x, top, x, bottom)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import patterns


class Numbered(object):
    __metaclass__ = patterns.NumberedInstances
    
    def __init__(self, instanceNumber=-1):
        self.instanceNumber = instanceNumber
    

class SubclassOfNumbered(Numbered):
    pass


class NumberedInstancesTestsMixin(object):
    ''' The tests below should work for a class with NumberedInstances as 
        metaclass as well as for a subclass of a class with NumberedInstances
        as metaclass. '''
        
    def testCounterIncreasesAfterEachInstantation(self):
        instances = []
        for count in range(3):
            instance = self.classUnderTest()
            self.assertEqual(count, instance.instanceNumber)
            instances.append(instance)
        
    def testInstanceNumbersAreReusedWhenFreed(self):
        instance1 = self.classUnderTest()
        del instance1
        instance2 = self.classUnderTest()
        self.assertEqual(0, instance2.instanceNumber)
        
    def testInstanceNumbersAreTheLowestFreeNumber(self):
        instance1 = self.classUnderTest()
        instance2 = self.classUnderTest()
        instance2Number = instance2.instanceNumber
        del instance2
        instance3 = self.classUnderTest()
        self.assertEqual(instance3.instanceNumber, instance2Number)
        
    def testInstanceNumbersFillTheGap(self):
        instances = []
        for count in range(10):
            instances.append(self.classUnderTest())
        del instances[4:6]
        instance4 = self.classUnderTest()
        self.assertEqual(4, instance4.instanceNumber)
        instance5 = self.classUnderTest()
        self.assertEqual(5, instance5.instanceNumber)
        instance10 = self.classUnderTest()
        self.assertEqual(10, instance10.instanceNumber)
        
        
class NumberedInstancesTest(NumberedInstancesTestsMixin, 
                            test.TestCase):
    classUnderTest = Numbered


class SubclassOfNumberedInstancesTest(NumberedInstancesTestsMixin, 
                                      test.TestCase):
    classUnderTest = SubclassOfNumbered

    def testSubclassInstancesHaveTheirOwnNumbers(self):
        numberedInstance = Numbered()
        subclassOfNumberedInstance = SubclassOfNumbered()
        self.assertEqual(0, numberedInstance.instanceNumber)
        self.assertEqual(0, subclassOfNumberedInstance.instanceNumber)





'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.syncml.tasksource import TaskSource
from taskcoachlib.syncml.notesource import NoteSource
from taskcoachlib.syncml.config import SyncMLConfigNode
from taskcoachlib.syncml.core import *

from taskcoachlib.i18n import _
from taskcoachlib.meta import data

import sys, wx


class AuthenticationFailure(Exception):
    pass


class TaskCoachManagementNode(ManagementNode):
    def __init__(self, syncMLConfig, *args, **kwargs):
        super(TaskCoachManagementNode, self).__init__(*args, **kwargs)

        self.__cfg = self.__getConfig(syncMLConfig)

    def __getConfig(self, cfg):
        for name in self.fullName.split('/'):
            for child in cfg.children():
                if child.name == name:
                    cfg = child
                    break
            else:
                child = SyncMLConfigNode(name)
                cfg.addChild(child)
                cfg = child

        return cfg

    def getChildrenMaxCount(self):
        return len(self.__cfg.children())

    def getChildrenNames(self):
        return [child.name for child in self.__cfg.children()]

    def readPropertyValue(self, name):
        return self.__cfg.get(name.decode('UTF-8')).encode('UTF-8')

    def setPropertyValue(self, name, value):
        self.__cfg.set(name.decode('UTF-8'), value.decode('UTF-8'))

    def clone(self):
        return self


class TaskCoachDMTree(DMTree):
    def __init__(self, syncMLConfig, *args, **kwargs):
        super(TaskCoachDMTree, self).__init__(*args, **kwargs)

        self.__syncMLConfig = syncMLConfig

    def isLeaf(self, node):
        return TaskCoachManagementNode(self.__syncMLConfig, node).getMaxChildrenCount() == 0

    def readManagementNode(self, nodeName):
        node = TaskCoachManagementNode(self.__syncMLConfig, nodeName)

        for name in node.getChildrenNames():
            node.addChild(TaskCoachManagementNode(self.__syncMLConfig, nodeName, name))

        return node


class TaskCoachDMTClientConfig(DMTClientConfig):
    def __init__(self, syncMLConfig, *args, **kwargs):
        super(TaskCoachDMTClientConfig, self).__init__(*args, **kwargs)

        self.__syncMLConfig = syncMLConfig

    def syncMLConfig(self):
        return self.__syncMLConfig

    def createDMTree(self, rootContext):
        return TaskCoachDMTree(self.__syncMLConfig, rootContext)


class Synchronizer(wx.ProgressDialog):
    def __init__(self, reportCallback, taskFile, password):
        super(Synchronizer, self).__init__(_('Synchronization'),
                                           _('Synchronizing. Please wait.\n\n\n'))

        self.clientName = 'TaskCoach-%s' % taskFile.guid().encode('UTF-8')
        self.reportCallback = reportCallback
        self.taskFile = taskFile

        cfg = taskFile.syncMLConfig()

        self.username = cfg[self.clientName]['spds']['syncml']['Auth'].get('username').encode('UTF-8') # Hum...
        self.password = password.encode('UTF-8')
        self.url = cfg[self.clientName]['spds']['syncml']['Conn'].get('syncUrl').encode('UTF-8')

        self.synctasks = cfg[self.clientName]['spds']['sources']['%s.Tasks' % self.clientName].get('dosync') == 'True'
        self.syncnotes = cfg[self.clientName]['spds']['sources']['%s.Notes' % self.clientName].get('dosync') == 'True'

        self.taskdbname = cfg[self.clientName]['spds']['sources']['%s.Tasks' % self.clientName].get('uri').encode('UTF-8')
        self.notedbname = cfg[self.clientName]['spds']['sources']['%s.Notes' % self.clientName].get('uri').encode('UTF-8')

        self.taskmode = cfg[self.clientName]['spds']['sources']['%s.Tasks' % self.clientName].get('preferredsyncmode')
        self.notemode = cfg[self.clientName]['spds']['sources']['%s.Notes' % self.clientName].get('preferredsyncmode')

    def init(self):
        self.dmt = TaskCoachDMTClientConfig(self.taskFile.syncMLConfig(), self.clientName)

        if not (self.dmt.read() and \
                self.dmt.deviceConfig.devID == self.clientName):
            self.dmt.setClientDefaults()

        ac = self.dmt.accessConfig
        ac.username = self.username
        ac.password = self.password

        ac.useProxy = 0
        ac.syncURL = self.url
        self.dmt.accessConfig = ac

        dc = self.dmt.deviceConfig
        dc.devID = self.clientName
        dc.devType = 'workstation'
        dc.manufacturerName = 'Task Coach developers'
        dc.modelName = sys.platform
        dc.firmwareVersion = '0.0'
        dc.softwareVersion = data.version
        self.dmt.deviceConfig = dc

        # Tasks source configuration

        self.sources = []

        if self.synctasks:
            try:
                cfg = self.dmt.getSyncSourceConfig('%s.Tasks' % self.clientName)
            except ValueError:
                cfg = SyncSourceConfig('%s.Tasks' % self.clientName)

            cfg.URI = self.taskdbname
            cfg.syncModes = 'two-way'
            cfg.supportedTypes = 'text/calendar'
            cfg.version = '1.0'

            self.dmt.setSyncSourceConfig(cfg)

            src = TaskSource(self,
                             self.taskFile.tasks(),
                             self.taskFile.categories(),
                             '%s.Tasks' % self.clientName, cfg)
            src.preferredSyncMode = globals()[self.taskmode]
            self.sources.append(src)

        if self.syncnotes:
            try:
                cfg = self.dmt.getSyncSourceConfig('%s.Notes' % self.clientName)
            except ValueError:
                cfg = SyncSourceConfig('%s.Notes' % self.clientName)

            cfg.URI = self.notedbname
            cfg.syncModes = 'two-way'
            cfg.supportedTypes = 'text/x-vnote:1.1'
            cfg.version = '1.0'

            self.dmt.setSyncSourceConfig(cfg)

            src = NoteSource(self,
                             self.taskFile.notes(),
                             self.taskFile.categories(),
                             # This is ugly and doesn't work for every configuration but well...
                             'text/plain' if self.url.endswith('rpc.php') else 'text/x-vnote',
                             '%s.Notes' % self.clientName, cfg)
            src.preferredSyncMode = globals()[self.notemode]
            self.sources.append(src)
    
    def onAddItem(self):
        self.added += 1
        self.pulse()

    def onUpdateItem(self):
        self.updated += 1
        self.pulse()

    def onDeleteItem(self):
        self.deleted += 1
        self.pulse()

    def pulse(self):
        msg = _('%d items added.\n%d items updated.\n%d items deleted.') % (self.added,
                                                                            self.updated,
                                                                            self.deleted)
        self.Pulse(msg)

    def error(self, code, msg):
        self.reportCallback(_('An error occurred in the synchronization.\nError code: %d; message: %s') \
                            % (code, msg))

    def synchronize(self):
        if not self.username:
            self.reportCallback(_('You must first edit your SyncML Settings, in Edit/SyncML preferences.'))
            return False

        self.Centre()
        self.Show()

        self.added = 0
        self.updated = 0
        self.deleted = 0

        self.taskFile.beginSync()
        try:
            self.init()

            client = SyncClient()
            client.sync(self.dmt, self.sources)

            code = client.report.lastErrorCode

            if code:
                if code == 401:
                    raise AuthenticationFailure()
                self.error(code, client.report.lastErrorMsg)

                # TODO: undo local modifications ?
                return False

            self.dmt.save()
        finally:
            self.taskFile.setSyncMLConfig(self.dmt.syncMLConfig())
            self.taskFile.endSync()

        return True



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import time, string, datetime, re, timedelta # pylint: disable=W0402
from taskcoachlib import patterns
from .fix import StrftimeFix


infinite = datetime.date.max

class RealDate(StrftimeFix, datetime.date):
    def __add__(self, delta):
        newdate = super(RealDate, self).__add__(delta)
        return RealDate(newdate.year, newdate.month, newdate.day)

    def __sub__(self, other):
        newdate = super(RealDate, self).__sub__(other)
        if isinstance(newdate, datetime.timedelta):
            return timedelta.TimeDelta(newdate.days, newdate.seconds, newdate.microseconds)
        else:
            return RealDate(newdate.year, newdate.month, newdate.day)


class InfiniteDate(datetime.date):
    __metaclass__ = patterns.Singleton

    def __new__(self):
        return super(InfiniteDate, self).__new__(InfiniteDate, infinite.year,
            infinite.month, infinite.day)

    def _getyear(self):
        return None

    year = property(_getyear)

    def _getmonth(self):
        return None

    month = property(_getmonth)

    def _getday(self):
        return None

    day = property(_getday)

    def __str__(self):
        return ''

    def __sub__(self, other):
        if isinstance(other, InfiniteDate):
            return timedelta.TimeDelta()
        else:
            return timedelta.TimeDelta.max

    def __add__(self, delta):
        # Whatever is added to InfiniteDate, we're still infinite:
        return self

# factories:

def parseDate(dateString, default=None):
    try:
        return Date(*[string.atoi(part) for part in dateString.split('-')])
    except ValueError:
        if default:
            return default
        else:
            return Date()

def Date(year=infinite.year, month=infinite.month, day=infinite.day):
    if (year, month, day) == (infinite.year, infinite.month, infinite.day):
        return InfiniteDate()
    else:
        return RealDate(year, month, day)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test, StringIO
from taskcoachlib import persistence, config
from taskcoachlib.domain import task, category, date


class TodoTxtReaderTestCase(test.TestCase):
    def setUp(self):
        task.Task.settings = config.Settings(load=False)
        self.tasks = task.TaskList()
        self.categories = category.CategoryList() 
        self.reader = persistence.TodoTxtReader(self.tasks, self.categories)
        
    def read(self, text, **kwargs):
        self.reader.readFile(StringIO.StringIO(text), **kwargs)
        
    def assertTaskSubject(self, *subjects):
        self.assertEqual(set(subjects), set(t.subject() for t in self.tasks))
        
    def assertCategorySubject(self, *subjects):
        self.assertEqual(set(subjects), set(c.subject() for c in self.categories))
        
    def assertPriority(self, priority):
        self.assertEqual(priority, list(self.tasks)[0].priority())
        
    def assertStartDate(self, *dateTimeArgs):
        self.assertEqual(date.DateTime(*dateTimeArgs), 
                         list(self.tasks)[0].plannedStartDateTime())
        
    def assertCompletionDate(self, *dateTimeArgs, **kwargs):
        expectedDateTime = kwargs['dateTime'] if 'dateTime' in kwargs else date.DateTime(*dateTimeArgs)
        self.assertEqual(expectedDateTime, list(self.tasks)[0].completionDateTime())
        
    def assertDueDate(self, *dateTimeArgs):
        self.assertEqual(date.DateTime(*dateTimeArgs), 
                         list(self.tasks)[0].dueDateTime())
        
    def assertTaskIsCompleted(self):
        self.failUnless(list(self.tasks)[0].completed())
        
    def testEmptyFile(self):
        self.reader.readFile(StringIO.StringIO())
        self.failIf(self.tasks)
        
    def testReadOneTask(self):
        self.read('Get milk\n')
        self.assertTaskSubject('Get milk')
        
    def testReadTwoTasks(self):
        self.read('Get milk\nDo laundry\n')
        self.assertTaskSubject('Get milk', 'Do laundry')
        
    def testTaskWithPriority(self):
        self.read('(A) Get milk\n')
        self.assertPriority(1)
        self.assertTaskSubject('Get milk')
        
    def testTaskWithStartDate(self):
        self.read('2011-01-31 Get milk\n')
        self.assertStartDate(2011, 1, 31)
        
    def testTaskWithStartDateWithoutLeadingZero(self):
        self.read('2011-1-31 Get milk\n')
        self.assertStartDate(2011, 1, 31)
        
    def testTaskWithPriorityAndStartDate(self):
        self.read('(Z) 2011-01-31 Get milk\n')
        self.assertPriority(26)
        self.assertStartDate(2011, 1, 31)
        
    def testCompletedTaskWithoutCompletionDate(self):
        now = date.Now()
        self.read('x Do dishes\n', now=lambda: now)
        self.assertTaskIsCompleted()
        self.assertCompletionDate(dateTime=now)
        
    def testCompletedTaskWithCompletionDate(self):
        self.read('X 2011-02-22 Do dishes\n')
        self.assertTaskIsCompleted()
        self.assertCompletionDate(2011, 2, 22)

    def testTaskWithStartAndCompletionDate(self):
        self.read('X 2011-2-22 2011-2-21 Do dishes\n')
        self.assertTaskIsCompleted()
        self.assertCompletionDate(2011, 2, 22)
        self.assertStartDate(2011, 2, 21)
    
    def testTaskWithSimpleContext(self):
        self.read('Order pizza @phone\n')
        self.assertCategorySubject(('@phone'))
        phone = list(self.categories)[0]
        pizza = list(self.tasks)[0]
        self.assertEqual(set([pizza]), phone.categorizables())
        self.assertEqual(set([phone]), pizza.categories())

    def testTaskWithSimpleProject(self):
        self.read('Order pizza +phone\n')
        self.assertCategorySubject(('+phone'))
        phone = list(self.categories)[0]
        pizza = list(self.tasks)[0]
        self.assertEqual(set([pizza]), phone.categorizables())
        self.assertEqual(set([phone]), pizza.categories())
        
    def testTaskWithPlusSign(self):
        self.read('Order pizza + drink\n')
        self.assertTaskSubject('Order pizza + drink')
        self.failIf(self.categories)
        
    def testTaskWithAtSign(self):
        self.read('Mail frank@niessink.com\n')
        self.assertTaskSubject('Mail frank@niessink.com')
        self.failIf(self.categories)
        
    def testTwoTasksWithTheSameContext(self):
        self.read('Order pizza @phone\nCall mom @phone\n')
        self.assertEqual(1, len(self.categories))
        phone = list(self.categories)[0]
        self.assertEqual(set(self.tasks), phone.categorizables())
        self.assertEqual([set([phone]), set([phone])], 
                         [t.categories() for t in self.tasks])
        
    def testTaskWithSubcategoryAsContext(self):
        self.read('Order pizza @home->phone\n')
        home = [c for c in self.categories if not c.parent()][0]
        self.assertEqual('@home', home.subject())
        phone = home.children()[0]
        self.assertEqual('phone', phone.subject())
        pizza = list(self.tasks)[0]
        self.assertEqual(set([pizza]), phone.categorizables())
        self.assertEqual(set([phone]), pizza.categories())
        
    def testTwoTasksWithTheSameSubcategory(self):
        self.read('Order pizza @home->phone\nOrder flowers @home->phone\n')
        home = [c for c in self.categories if not c.parent()][0]
        self.assertEqual('@home', home.subject())
        phone = home.children()[0]
        self.assertEqual('phone', phone.subject())
        self.assertEqual(set(self.tasks), phone.categorizables())
        for eachTask in self.tasks:
            self.assertEqual(set([phone]), eachTask.categories())
            
    def testTaskWithMultipleContexts(self):
        self.read('Order pizza @phone @food\n')
        self.assertEqual(2, len(self.categories))
        pizza = list(self.tasks)[0]
        self.assertEqual(set(self.categories), pizza.categories())
        
    def testContextBeforeTask(self):
        self.read('@phone Order pizza\n')
        self.assertEqual(1, len(self.categories))

    def testProjectBeforeTask(self):
        self.read('+phone Order pizza\n')
        self.assertEqual(1, len(self.categories))
        
    def testPriorityAndContextBeforeTask(self):
        self.read('(A) @phone thank Mom for the meatballs')
        self.assertEqual(1, len(self.categories))
        self.assertCategorySubject('@phone')
        phone = list(self.categories)[0]
        thankMom = list(self.tasks)[0]
        self.assertEqual(set([thankMom]), phone.categorizables())
        self.assertEqual(set([phone]), thankMom.categories())
                
    def testPriorityAndProjectAndContextBeforeTask(self):
        self.read('(B) +GarageSale @phone schedule Goodwill pickup')
        self.assertEqual(2, len(self.categories))
        
    def testAutomaticallyCreateParentTask(self):
        self.read('Project->Activity')
        self.assertEqual(2, len(self.tasks))
        parent = [t for t in self.tasks if not t.parent()][0]
        self.assertEqual('Project', parent.subject())
        self.assertEqual('Activity', parent.children()[0].subject())
        
    def testAutomaticallyCreateParentTask_WithSpaces(self):
        self.read('Project -> Activity')
        self.assertEqual(2, len(self.tasks))
        parent = [t for t in self.tasks if not t.parent()][0]
        self.assertEqual('Project', parent.subject())
        self.assertEqual('Activity', parent.children()[0].subject())
        
    def testFirstChildAndThenParent(self):
        self.read('Project->Activity\nProject\n')
        self.assertEqual(2, len(self.tasks))
        parent = [t for t in self.tasks if not t.parent()][0]
        self.assertEqual('Project', parent.subject())
        self.assertEqual('Activity', parent.children()[0].subject())
        
    def testIgnoreEmptyLine(self):
        self.read('\n')
        self.failIf(self.tasks)
        
    def testDueDate(self):
        self.read('Import due date due:2011-03-05\n')
        self.assertEqual('Import due date', list(self.tasks)[0].subject())
        self.assertEqual(date.DateTime(2011, 3, 5, 0, 0, 0), 
                         list(self.tasks)[0].dueDateTime())

    def testDueDateBeforeProject(self):
        self.read('Import due date due:2011-03-05 +TaskCoach\n')
        self.assertTaskSubject('Import due date')
        self.assertDueDate(2011, 3, 5)
        self.assertCategorySubject('+TaskCoach')
        
    def testTrailingJunk(self):
        taskWithJunk = 'Test +this_is_not_a_project due to trailing junk'
        self.read(taskWithJunk)
        self.assertTaskSubject(taskWithJunk)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import operating_system
from taskcoachlib.i18n import _
from taskcoachlib.thirdparty import aui
import notebook
import wx
import wx.html
from wx.lib import sized_controls
import os


class Dialog(sized_controls.SizedDialog):
    def __init__(self, parent, title, bitmap='edit', 
                 direction=None, *args, **kwargs):
        self._buttonTypes = kwargs.get('buttonTypes', wx.OK | wx.CANCEL)
        super(Dialog, self).__init__(parent, -1, title,
            style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER | wx.MAXIMIZE_BOX | wx.MINIMIZE_BOX)
        self.SetIcon(wx.ArtProvider_GetIcon(bitmap, wx.ART_FRAME_ICON,
            (16, 16)))

        if operating_system.isWindows7_OrNewer():
            # Without this the window has no taskbar icon on Windows, and the focus comes back to the main
            # window instead of this one when returning to Task Coach through Alt+Tab. Which is probably not
            # what we want.
            import win32gui, win32con
            exStyle = win32gui.GetWindowLong(self.GetHandle(), win32con.GWL_EXSTYLE)
            win32gui.SetWindowLong(self.GetHandle(), win32con.GWL_EXSTYLE, exStyle|win32con.WS_EX_APPWINDOW)

        self._panel = self.GetContentsPane()
        self._panel.SetSizerType('vertical')
        self._panel.SetSizerProps(expand=True, proportion=1)
        self._direction = direction
        self._interior = self.createInterior()
        self._interior.SetSizerProps(expand=True, proportion=1)
        self.fillInterior()
        self._buttons = self.createButtons()
        self._panel.Fit()
        self.Fit()
        self.CentreOnParent()
        if not operating_system.isGTK():
            wx.CallAfter(self.Raise)
        wx.CallAfter(self._panel.SetFocus)

    def SetExtraStyle(self, exstyle):
        # SizedDialog's constructor calls this to set WS_EX_VALIDATE_RECURSIVELY. We don't need
        # it, it makes the dialog appear in about 7 seconds, and it makes switching focus
        # between two controls take up to 5 seconds.
        pass

    def createInterior(self):
        raise NotImplementedError

    def fillInterior(self):
        pass
    
    def createButtons(self):
        buttonTypes = wx.OK if self._buttonTypes == wx.ID_CLOSE else self._buttonTypes
        buttonSizer = self.CreateStdDialogButtonSizer(buttonTypes)
        if self._buttonTypes & wx.OK or self._buttonTypes & wx.ID_CLOSE:
            buttonSizer.GetAffirmativeButton().Bind(wx.EVT_BUTTON, self.ok)
        if self._buttonTypes & wx.CANCEL:
            buttonSizer.GetCancelButton().Bind(wx.EVT_BUTTON, self.cancel)
        if self._buttonTypes == wx.ID_CLOSE:
            buttonSizer.GetAffirmativeButton().SetLabel(_('Close'))
        self.SetButtonSizer(buttonSizer)
        return buttonSizer
    
    def ok(self, event=None):
        if event:
            event.Skip()
        self.Close(True)
        self.Destroy()
        
    def cancel(self, event=None):
        if event:
            event.Skip()
        self.Close(True)
        self.Destroy()
        
    def disableOK(self):
        self._buttons.GetAffirmativeButton().Disable()
        
    def enableOK(self):
        self._buttons.GetAffirmativeButton().Enable()


class NotebookDialog(Dialog):    
    def createInterior(self):
        return notebook.Notebook(self._panel, 
            agwStyle=aui.AUI_NB_DEFAULT_STYLE & ~aui.AUI_NB_TAB_SPLIT & \
                     ~aui.AUI_NB_TAB_MOVE & ~aui.AUI_NB_DRAW_DND_TAB)

    def fillInterior(self):
        self.addPages()
            
    def __getitem__(self, index):
        return self._interior[index]
    
    def ok(self, *args, **kwargs):
        self.okPages()
        super(NotebookDialog, self).ok(*args, **kwargs)
        
    def okPages(self, *args, **kwargs):
        for page in self._interior:
            page.ok(*args, **kwargs)

    def addPages(self):
        raise NotImplementedError 

        
class HtmlWindowThatUsesWebBrowserForExternalLinks(wx.html.HtmlWindow):
    def OnLinkClicked(self, linkInfo):  # pylint: disable=W0221
        openedLinkInExternalBrowser = False
        if linkInfo.GetTarget() == '_blank':
            import webbrowser  # pylint: disable=W0404
            try:
                webbrowser.open(linkInfo.GetHref())
                openedLinkInExternalBrowser = True
            except webbrowser.Error:
                pass
        if not openedLinkInExternalBrowser:
            super(HtmlWindowThatUsesWebBrowserForExternalLinks, 
                  self).OnLinkClicked(linkInfo)


class HTMLDialog(Dialog):
    def __init__(self, title, htmlText, *args, **kwargs):
        self._htmlText = htmlText
        super(HTMLDialog, self).__init__(None, title, buttonTypes=wx.ID_CLOSE, 
                                         *args, **kwargs)
        
    def createInterior(self):
        interior = HtmlWindowThatUsesWebBrowserForExternalLinks(self._panel, 
            -1, size=(550, 400))
        if self._direction:
            interior.SetLayoutDirection(self._direction)
        return interior
        
    def fillInterior(self):
        self._interior.AppendToPage(self._htmlText)

    def OnLinkClicked(self, linkInfo):
        pass
        
        
def AttachmentSelector(**callerKeywordArguments):
    kwargs = {'message': _('Add attachment'),
              'default_path': os.getcwd(), 
              'wildcard': _('All files (*.*)|*'), 
              'flags': wx.FD_OPEN}
    kwargs.update(callerKeywordArguments)
    return wx.FileSelector(**kwargs)  # pylint: disable=W0142



# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test, StringIO, time
from taskcoachlib import persistence, gui, config, meta
from taskcoachlib.domain import task, effort, date


class UTF8StringIO(StringIO.StringIO):
    """
    Mimic codecs.open; encodes on the fly in UTF-8
    """
    def write(self, u):
        StringIO.StringIO.write(self, u.encode('UTF-8'))


class VCalTestCase(test.wxTestCase):
    selectionOnly = 'Subclass responsibility'
    
    def setUp(self):
        super(VCalTestCase, self).setUp()
        task.Task.settings = self.settings = config.Settings(load=False)
        self.fd = UTF8StringIO()
        self.writer = persistence.iCalendarWriter(self.fd)
        self.taskFile = persistence.TaskFile()

    def tearDown(self):
        super(VCalTestCase, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    def writeAndRead(self):
        self.writer.write(self.viewer, self.settings, self.selectionOnly)
        return self.fd.getvalue().decode('utf-8')

    def selectItems(self, items):
        self.viewer.select(items)
        
    def numberOfSelectedItems(self):
        return len(self.viewer.curselection())
    
    def numberOfVisibleItems(self):
        return self.viewer.size()
    
    
class VCalendarCommonTestsMixin(object):        
    def testStart(self):
        self.assertEqual('BEGIN:VCALENDAR', self.vcalFile.split('\r\n')[0])
        
    def testVersion(self):
        self.assertEqual('VERSION:2.0', self.vcalFile.split('\r\n')[1])

    def testProdId(self):
        domain = meta.url[len('http://'):-1]
        self.assertEqual('PRODID:-//%s//NONSGML %s V%s//EN'%(domain,
            meta.name, meta.version), self.vcalFile.split('\r\n')[2])

    def testEnd(self):
        self.assertEqual('END:VCALENDAR', self.vcalFile.split('\r\n')[-2])


class VCalEffortWriterTestCase(VCalTestCase):
    def setUp(self):
        super(VCalEffortWriterTestCase, self).setUp()        
        self.task1 = task.Task(u'Task 1')
        self.effort1 = effort.Effort(self.task1, description=u'Description',
                                     start=date.DateTime(2000,1,1,1,1,1),
                                     stop=date.DateTime(2000,2,2,2,2,2))
        self.effort2 = effort.Effort(self.task1)
        self.task1.addEffort(self.effort1)
        self.task1.addEffort(self.effort2)
        self.taskFile.tasks().extend([self.task1])
        self.viewer = gui.viewer.EffortViewer(self.frame, self.taskFile,
                                              self.settings)
        self.viewer.widget.select([self.effort1])
        self.viewer.updateSelection()
        self.vcalFile = self.writeAndRead()


class VCalEffortCommonTestsMixin(VCalendarCommonTestsMixin):
    def testBeginVEvent(self):
        self.assertEqual(self.expectedNumberOfItems(), 
                         self.vcalFile.count('BEGIN:VEVENT'))

    def testEndVEvent(self):
        self.assertEqual(self.expectedNumberOfItems(), 
                         self.vcalFile.count('END:VEVENT'))
        
    def testEffortSubject(self):
        self.failUnless(u'SUMMARY:Task 1' in self.vcalFile)

    def testEffortDescription(self):
        self.failUnless(u'DESCRIPTION:Description' in self.vcalFile)
        
    def testEffortStart(self):
        startLocal = date.DateTime(2000, 1, 1, 1, 1, 1)
        startUTC = startLocal.utcfromtimestamp(time.mktime(startLocal.timetuple()))
        self.failUnless('DTSTART:%04d%02d%02dT%02d%02d%02dZ' % (startUTC.year,
                                                                startUTC.month,
                                                                startUTC.day,
                                                                startUTC.hour,
                                                                startUTC.minute,
                                                                startUTC.second))

    def testEffortEnd(self):
        endLocal = date.DateTime(2000, 2, 2, 2, 2, 2)
        endUTC = endLocal.utcfromtimestamp(time.mktime(endLocal.timetuple()))
        self.failUnless('DTEND:%04d%02d%02dT%02d%02d%02dZ' % (endUTC.year,
                                                              endUTC.month,
                                                              endUTC.day,
                                                              endUTC.hour,
                                                              endUTC.minute,
                                                              endUTC.second))
        
    def testEffortId(self):
        self.failUnless('UID:%s'%self.effort1.id() in self.vcalFile)

    
class VCalEffortWriterTest(VCalEffortWriterTestCase,
                           VCalEffortCommonTestsMixin):
    selectionOnly = False
    
    def expectedNumberOfItems(self):
        return self.numberOfVisibleItems()
        

class VCalEffortWriterSelectionOnlyTest(VCalEffortWriterTestCase,
                                        VCalEffortCommonTestsMixin):
    selectionOnly = True

    def expectedNumberOfItems(self):
        return self.numberOfSelectedItems()
            

class VCalTaskWriterTestCase(VCalTestCase):
    treeMode = 'Subclass responsibility'
    
    def setUp(self):
        super(VCalTaskWriterTestCase, self).setUp() 
        self.task1 = task.Task(u'Task subject 1', 
                               description='Task description 1',
                               percentageComplete=56,
                               creationDateTime=date.DateTime.min)
        self.task2 = task.Task(u'Task subject 2黑', 
                               description=u'Task description 2\nwith newline\n微软雅黑',
                               modificationDateTime=date.DateTime(2012, 1, 1))
        self.taskFile.tasks().extend([self.task1, self.task2])
        self.settings.set('taskviewer', 'treemode', self.treeMode)
        self.viewer = gui.viewer.TaskViewer(self.frame, self.taskFile,
            self.settings)
        self.selectItems([self.task2])
        self.vcalFile = self.writeAndRead()
        
        
class VCalTaskCommonTestsMixin(VCalendarCommonTestsMixin):
    def testTaskSubject(self):
        self.failUnless(u'SUMMARY:Task subject 2' in self.vcalFile)
        
    def testTaskDescription(self):
        self.failUnless(u'DESCRIPTION:Task description 2\r\n with newline\r\n 微软雅黑' in self.vcalFile, self.vcalFile)

    def testNumber(self):
        self.assertEqual(self.expectedNumberOfItems(),
                         self.vcalFile.count('BEGIN:VTODO'))  # pylint: disable=W0511

    def testTaskId(self):
        self.failUnless('UID:%s' % self.task2.id() in self.vcalFile)
        
    def testCreationDateTime(self):
        creation_datetime = persistence.icalendar.ical.fmtDateTime(self.task2.creationDateTime())
        self.failUnless('CREATED:%s' % creation_datetime in self.vcalFile)
        
    def testMissingCreationDateTime(self):
        self.assertEqual(1, self.vcalFile.count('CREATED:'))
        
    def testModificationDateTime(self):
        modification_datetime = persistence.icalendar.ical.fmtDateTime(date.DateTime(2012, 1, 1))
        self.failUnless('LAST-MODIFIED:%s' % modification_datetime in self.vcalFile)
        
    def testMissingModificationDateTime(self):
        self.assertEqual(1, self.vcalFile.count('LAST-MODIFIED'))


class TestSelectionOnlyMixin(VCalTaskCommonTestsMixin):
    selectionOnly = True

    def expectedNumberOfItems(self):
        return self.numberOfSelectedItems()  


class TestSelectionList(TestSelectionOnlyMixin, VCalTaskWriterTestCase):
    treeMode = 'False'

class TestSelectionTree(TestSelectionOnlyMixin, VCalTaskWriterTestCase):
    treeMode = 'True'


class TestNotSelectionOnlyMixin(VCalTaskCommonTestsMixin):
    selectionOnly = False

    def expectedNumberOfItems(self):
        return self.numberOfVisibleItems()

    def testPercentageComplete(self):
        self.failUnless('PERCENT-COMPLETE:56' in self.vcalFile)


class TestNotSelectionList(TestNotSelectionOnlyMixin, VCalTaskWriterTestCase):
    treeMode = 'False'

class TestNotSelectionTree(TestNotSelectionOnlyMixin, VCalTaskWriterTestCase):
    treeMode = 'True'
    

class FoldTest(test.TestCase):
    def setUp(self):
        super(FoldTest, self).setUp()
        self.fold = persistence.icalendar.ical.fold

    def testEmptyText(self):
        self.assertEqual('', self.fold([]))
        
    def testDontFoldAShortLine(self):
        self.assertEqual('Short line\r\n', self.fold(['Short line']))
        
    def testFoldALongLine(self):
        self.assertEqual('Long \r\n line\r\n', self.fold(['Long line'], 
                                                         linewidth=5))
        
    def testFoldAReallyLongLine(self):
        self.assertEqual('Long\r\n  li\r\n ne\r\n', self.fold(['Long line'], 
                                                              linewidth=4))
        
    def testFoldTwoShortLines(self):
        self.assertEqual('Short line\r\n'*2, self.fold(['Short line']*2))
        
    def testFoldTwoLongLines(self):
        self.assertEqual('Long \r\n line\r\n'*2, self.fold(['Long line']*2, 
                                                         linewidth=5))

    def testFoldALineWithNewLines(self):
        self.assertEqual('Line 1\r\n Line 2\r\n', self.fold(['Line 1\nLine 2']))
        



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import sys

def log_call(traceback_depth):
    ''' Decorator for function calls that prints the function name,
        arguments and result to stdout. Usage:
        
        @log_call(traceback_depth)
        def function(arg):
            ...
    '''
    # Import here instead of at the module level to prevent unnecessary 
    # inclusion of the inspect module when packaging the application:
    import inspect 
    
    def outer(func):        
        def inner(*args, **kwargs):
            result = func(*args, **kwargs)
            write = sys.stdout.write
            for frame in inspect.stack(context=2)[traceback_depth:0:-1]:
                write(format_traceback(frame))
            write('%s\n'%signature(func, args, kwargs, result))
            write('===\n')
            return result
        return inner
    return outer    
        
    
def time_call(func):
    ''' Decorator for function calls that times the call. '''
    
    import time
    
    def inner(*args, **kwargs):
        start = time.time() 
        result = func(*args, **kwargs)
        stop = time.time()
        sys.stdout.write('%s took %f seconds\n'%\
            (signature(func, args, kwargs, result), stop-start))
        return result
    return inner


def profile_call(func):
    ''' Docorator for profiling a specific function. I'm not sure what
        happens if you decorate a recursive function... '''

    import hotshot
    
    def inner(*args, **kwargs):
        profiler = hotshot.Profile('.profile')
        return profiler.runcall(func, *args, **kwargs)
    return inner


def signature(func, args, kwargs, result):
    func = func.__name__
    result = unicode(result)
    try:
        return '%s(%s, %s) -> %s'%(func, unicode(args), unicode(kwargs), result)
    except:
        return '%s(...) -> %s'%(func, result) # pylint: disable=W0702
                               
                               
def format_traceback(frame):
    result = []
    filename, lineno, caller, context = frame[1:5]
    result.append('  File "%s", line %s, in %s'%(filename, lineno, caller))
    for line in context:
        result.append(line[:-1])
    return '\n'.join(result) + '\n'   
    



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import gui, config, persistence
from taskcoachlib.domain import task, effort, date
from unittests import dummy


class EditorUnderTest(gui.dialog.editor.EffortEditor):        
    def __init__(self, *args, **kwargs):
        super(EditorUnderTest, self).__init__(*args, **kwargs)
        self.editorClosed = False
                
    def onClose(self, event): # pragma: no cover
        self.editorClosed = True
        super(EditorUnderTest, self).onClose(event)
        
        
class EffortEditorTest(test.wxTestCase):      
    def setUp(self):
        super(EffortEditorTest, self).setUp()
        task.Task.settings = self.settings = config.Settings(load=False)
        self.taskFile = persistence.TaskFile()
        self.taskList = self.taskFile.tasks()
        self.effortList = self.taskFile.efforts()
        self.task = task.Task('task')
        self.effort = effort.Effort(self.task)
        self.task.addEffort(self.effort)
        self.task2 = task.Task('task2')
        self.taskFile.tasks().extend([self.task, self.task2])
        self.editor = EditorUnderTest(self.frame, 
            list(self.effortList), self.settings, self.taskFile.efforts(), 
            self.taskFile, raiseDialog=False)

    def tearDown(self):
        super(EffortEditorTest, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    def createEditor(self):
        return EditorUnderTest(self.frame,
            list(self.taskFile.efforts()), self.settings, 
            self.taskFile.efforts(), self.taskFile)        

    # pylint: disable=W0201,W0212
        
    def testCreate(self):
        self.assertEqual(self.task, self.editor._interior._taskEntry.GetValue())
        self.assertEqual(self.effort.getStart().date(), 
            self.editor._interior._startDateTimeEntry.GetValue().date())
        self.assertEqual(self.effort.task(), 
            self.editor._interior._taskEntry.GetValue())    
        
    def testInvalidEffort(self):    
        self.editor._interior._stopDateTimeEntry.SetValue(date.DateTime(1900, 1, 1))
        self.editor._interior.onDateTimeChanged(dummy.Event())
        self.failUnless(self.editor._interior._invalidPeriodMessage.GetLabel())
        
    def testChangeTask(self):
        self.editor._interior._taskEntry.SetValue(self.task2)
        self.editor._interior._taskSync.onAttributeEdited(dummy.Event())
        self.assertEqual(self.task2, self.effort.task())
        self.failIf(self.effort in self.task.efforts())
        
    def testChangeTaskDoesNotCloseEditor(self):
        self.editor._interior._taskEntry.SetValue(self.task2)
        self.editor._interior._taskSync.onAttributeEdited(dummy.Event())
        self.failIf(self.editor.editorClosed)
